#!/usr/bin/env python
#
# Glasnost
# By: Odile Bnassy <obenassy@entrouvert.com>
#     Thierry Dulieu <tdulieu@easter-eggs.com>
#     Frdric Pters <fpeters@theridion.com>
#     Benjamin Poussin <poussin@codelutin.com>
#     Emmanuel Raviart <eraviart@entrouvert.com>
#     Emmanuel Saracco <esaracco@easter-eggs.com>
#
# Copyright (C) 2000, 2001 Easter-eggs & Emmanuel Raviart
# Copyright (C) 2002 Odile Bnassy, Code Lutin, Thierry Dulieu, Easter-eggs,
#     Entr'ouvert, Frdric Pters, Benjamin Poussin, Emmanuel Raviart,
#     Emmanuel Saracco & Thridion
# Copyright (C) 2003 Odile Bnassy, Code Lutin, Thierry Dulieu, Easter-eggs,
#     Entr'ouvert, Ouvaton, Frdric Pters, Benjamin Poussin, Rodolphe
#     Quideville, Emmanuel Raviart, Emmanuel Saracco, Thridion & Vecam
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


from __future__ import nested_scopes
    # lambdas are so much fun with nested scopes...


__doc__ = """Glasnost Accounts Server"""

__version__ = '$Revision: 1.32 $'[11:-2]


import __builtin__
__builtin__.__dict__['_'] = lambda x: x
__builtin__.__dict__['N_'] = lambda x: x

import time

import glasnost as glasnost # Do not edit or remove this line!

from glasnost.common.AccountsCommon import *
import glasnost.common.faults as faults
import glasnost.common.context as context
from glasnost.common.tools import extractLocalId
import glasnost.common.xhtmlgenerator as X

from glasnost.server.ObjectsServer import register, ObjectServerMixin, \
     AdminServerMixin, ObjectsServer
from glasnost.server.tools import *

from glasnost.proxy.DispatcherProxy import MultiCall
from glasnost.proxy.tools import getProxyForServerRole

applicationName = 'AccountsServer'
applicationRole = 'accounts'
applicationTokens = None
dispatcher = None


class AdminAccounts(AdminServerMixin, AdminAccountsCommon):
    pass
register(AdminAccounts)


class Account(ObjectServerMixin, AccountCommon):
    def acquireNonCore(self, objectDirectoryPath = None,
                       dataDirectoryPath = None, parentSlot = None):
        ObjectServerMixin.acquireNonCore(
            self, objectDirectoryPath = objectDirectoryPath,
            dataDirectoryPath = dataDirectoryPath, parentSlot = parentSlot)
        if self.getServer().port is None:
            # The server is not running, so assume that Glasnost
            # is not running either (it happens for example during
            # ids conversion). Don't call the other servers. Do nothing.
            return        
        if self.ledgerId is None:
            self.getSlot('currency').setValue(None)
            self.getSlot('startDate').setValue(None)
        else:
            try:
                ledger = getProxyForServerRole('ledgers').getPartialObject(
                    self.ledgerId, ['currency', 'startDate'])
            except faults.MissingItem:
                self.getSlot('currency').setValue(None)
                self.getSlot('startDate').setValue(None)
            else:
                self.getSlot('currency').setValue(
                    ledger.getSlot('currency').getValue())
                self.getSlot('startDate').setValue(
                    ledger.getSlot('startDate').getValue())

    def canBeGottenByClient(self, clientId):
        clientNameAndPort, clientRole = splitApplicationId(clientId)
        return ObjectServerMixin.canBeGottenByClient(self, clientId) or \
                clientRole in ['balances','extracts','ledgers','records']

    def compute(self):
        membersProxy = getProxyForServerRole('records')
        timeNow = time.time()
        try:
            memberIds = membersProxy.searchObjectIds(self.id)
            memberObjects = [ membersProxy.getPartialObject(memberId, [
                'amount', 'certainty', 'date']) for memberId in memberIds ]
            members = [(
                int(extractLocalId(x.id)),
                int(x.date),
                x.certainty,
                x.amount
                ) for x in memberObjects ] 
        except:
            raise Exception(
                'Cannot get member objects for account %s' % self.id)
        try:
            balance = 0
            for member in members:
                if member[1] > timeNow:
                    break
                if member[2] >= 4:
                    balance += member[3]
        except:
            raise Exception('Cannot compute account %s balance' % self.id)
        self.getSlot('balance').setValue(balance)
        self.getSlot('members').setValue(members)

    def getBalance(self, dates, certaintyLevel):
        result = {}
        for date in dates:
            result[date] = 0
        for member in self.members:
            memberId, memberDate, memberCertainty, memberAmount = member
            for date in dates:
                if memberDate < date and memberCertainty >= certaintyLevel:
                    result[date] += memberAmount
        return result
                
    def modify(self, changes = None, givenSlotNames = None):
        ObjectServerMixin.modify(
            self, changes, givenSlotNames = givenSlotNames)
        self.compute()
        
    def releaseNonCore(self, parentSlot = None):
        ObjectServerMixin.releaseNonCore(self, parentSlot = parentSlot)
register(Account)


class AccountsServer(ObjectsServer):
    adminClassName = 'AdminAccounts'
    objectClassName = 'Account'
    version = __version__

    def computeObject(self, virtualServerId, clientToken, userToken, objectId):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            virtualServer = self.getVirtualServer(virtualServerId)
            if not virtualServer.objects.has_key(objectId):
                raise faults.MissingItem(objectId)
            object = virtualServer.objects[objectId]
            object.compute()
            self.saveVirtualServer(virtualServer)
            return objectId # We have to return something.
        finally:
            self.pullContext()
                    
    def registerXmlRpcMethods(self):
        ObjectsServer.registerXmlRpcMethods(self)
        self.registerXmlRpcMethod('computeObject')

    def repairVirtualServer(self, virtualServer, version):
        changed = 0
        if version < 6000:
            for id, object in virtualServer.objects.items():
                setattr(object, 'balance', 0)
                setattr(object, members, [])
                object.compute()
                for attr in [
                    'balanceAmount', 'balanceSide',
                    'initAmount', 'initSide',
                    'membersSet']:
                    delattr(object)
            changed = 1
        if version < 5004:
            for id, object in virtualServer.objects.items():
                for attr in ['balanceAmount', 'initAmount']: 
                    if not hasattr(object, attr):
                        setattr(object, attr, 0)
                        changed = 1
                for attr in [
                    'balanceSide', 'initSide', 'currency', 'ledgerId',
                    'membersSet', 'startDate']: 
                    if not hasattr(object, attr):
                        setattr(object, attr, None)
                        changed = 1        
            changed = virtualServer.admin.repair(5004) or changed
            for id, object in virtualServer.objects.items():
                changed = object.repair(5004) or changed
        if version <= 1011000:
            admin = virtualServer.admin
            if admin.id is None:
                changed = 1
                admin.id = '%s/__admin__' % virtualServer.virtualServerId
        if changed:
            self.saveVirtualServer(virtualServer)
            
accountsServer = AccountsServer()


if __name__ == "__main__":
    accountsServer.launch(applicationName, applicationRole)
