#!/usr/bin/env python
#
# Glasnost
# By: Odile Bnassy <obenassy@entrouvert.com>
#     Thierry Dulieu <tdulieu@easter-eggs.com>
#     Frdric Pters <fpeters@theridion.com>
#     Benjamin Poussin <poussin@codelutin.com>
#     Emmanuel Raviart <eraviart@entrouvert.com>
#     Emmanuel Saracco <esaracco@easter-eggs.com>
#
# Copyright (C) 2000, 2001 Easter-eggs & Emmanuel Raviart
# Copyright (C) 2002, 2003 Odile Bnassy, Code Lutin, Thierry Dulieu,
#     Easter-eggs, Entr'ouvert, Frdric Pters, Benjamin Poussin,
#     Emmanuel Raviart, Emmanuel Saracco & Thridion
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


__doc__ = """Glasnost Contacts Server"""

__version__ = '$Revision: 1.10 $'[11:-2]


import __builtin__
__builtin__.__dict__['_'] = lambda x: x
__builtin__.__dict__['N_'] = lambda x: x

import copy
import os
import time

import glasnost as glasnost # Do not edit or remove this line!

from glasnost.common.ContactsCommon import *
import glasnost.common.faults as faults
import glasnost.common.context as context
from glasnost.common.tools import splitId
import glasnost.common.xhtmlgenerator as X

from glasnost.server.ObjectsServer import register, ObjectServerMixin, AdminServerMixin, ObjectsServer, VirtualServer
from glasnost.server.tools import *

from glasnost.proxy.tools import getProxy

try:
    import ldap
except ImportError:
    #import sys
    #sys.stderr.write('Python-LDAP is not installed.  Not launching ContactsServer.\n')
    sys.exit(1)


applicationName = 'ContactsServer'
applicationRole = 'contacts'
applicationTokens = None
dispatcher = None

class AdminContacts(AdminServerMixin, AdminContactsCommon):
    pass
register(AdminContacts)


class Contact(ObjectServerMixin, ContactCommon):
    def saveNonCore(self, objectDirectoryPath = None, dataDirectoryPath = None,
                    parentSlot = None):
        ObjectServerMixin.saveNonCore(self, objectDirectoryPath,
                                      dataDirectoryPath, parentSlot)
        ## Save in LDAP
        server = self.getServer()
        virtualServerId = server.computeVirtualServerId(self.id)
        virtualServer = server.getVirtualServer(virtualServerId)
        virtualServer.objects[self.id] = self
register(Contact)

class LdapPersistence:
    def __init__(self, config, serverId, objectClass = None):
        self.config = config
        self.serverId = serverId
        self.config.objectClass = objectClass
        
    def connect(self):
        """
        return connection to the LDAP server. The admin user is bind
        """
        con = ldap.open(self.config.ldap_host, self.config.ldap_port)
        con.simple_bind_s(self.config.admin_dn, self.config.admin_password)
        return con

    def keys(self):
        return [o.getId() for o in self.values()]
    
    def items(self):
        return [(o.getId(), o) for o in self.values()]
    
    def values(self):
        return self.search(self.config.ldap_idField + '=*')
    
    def has_key(self, key):
        return self.__contains__(key)
    
    def get(self, key, failobj=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return failobj

    def search(self, searchString):
        if not self.config.objectClass:
            return []
        
        con = self.connect()
        data = con.search_s(self.config.suffix,
                            ldap.SCOPE_SUBTREE,
                            searchString,
                            None)
        con.unbind_s()

        result = []
        for key, value in data:
            o = self.config.objectClass()
            o.load(value)
            o.updateId(self.serverId)
            result.append(o)

        return result
      

    def __contains__(self, key):
        id = splitId(key)[-1]
        con = self.connect()
        result  = con.search_s(self.config.suffix,
                               ldap.SCOPE_SUBTREE,
                               self.config.ldap_idField + '=' + id,
                               None)        
        con.unbind_s()
        return len(result)

    def __delitem__(self, key):
        id = splitId(key)[-1]
        con = self.connect()
        con.delete_s(self.config.dn % id)
        con.unbind_s()

    def __getitem__(self, key):
        id = splitId(key)[-1]
        result = self.search(self.config.ldap_idField + '=' + id)
        return result[0]

    def __setitem__(self, key, value):
        ## the first time that object is added, we keep the object's class
        if not self.config.objectClass or \
               self.config.objectClass != value.__class__:
            self.config.objectClass = value.__class__

        id = splitId(key)[-1]

        data = value.storeList()
        con = self.connect()
        ## None key indicate a new Object
        try:
            if id == 'None':
                value.updateId(self.serverId)
                key = value.getId()
                id = splitId(key)[-1]
                data = [('objectClass', self.config.objectClassSet)] + data
                con.add_s(self.config.dn % id, data)
            else:
                data = [(ldap.MOD_REPLACE, name, value) for name, value in data]
                con.modify_s(self.config.dn % id, data)
            con.unbind_s()
        except:
            print 'Error during store: %s' % data
            raise

class LdapVirtualServer(VirtualServer):
    admin = None
    objects = None

    def init(self, server, virtualServerId):
        VirtualServer.init(self, server, virtualServerId)
        self.admin = server.newAdmin()
        if not virtualServerId:
            virtualServerId = '%s/%s'
        self.admin.id = '%s/__admin__' % virtualServerId
        self.objects = LdapPersistence(self.admin, virtualServerId)

    def getNextLocalId(self):
        return None


class ContactsVirtualServer(LdapVirtualServer):
    pass

class ContactsServer(ObjectsServer):
    adminClassName = 'AdminContacts'
    objectClassName = 'Contact'
    VirtualServer = ContactsVirtualServer
    version = __version__

    def registerXmlRpcMethods(self):
        ObjectsServer.registerXmlRpcMethods(self)
        self.registerXmlRpcMethod('findObjectIds')

    def convertSearchString(self, searchString):
        mapping = Contact().mapping
        for fieldName, ldapName in mapping:
            searchString = searchString.replace('%s=' % fieldName, '%s=' % ldapName)
        return searchString
        
    def findObjectIds(self, virtualServerId, clientToken, userToken,
                      searchString = '', sortedOrder = 'None'):
        """
        @param searchString: text for ldap request.
        @type searchString: String
        """
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            virtualServer = self.getVirtualServer(virtualServerId)
            if searchString:
                searchString = self.convertSearchString(searchString)
                foundIds = virtualServer.objects.search(searchString)
            else:
                foundIds = virtualServer.objects.values()

            #sort
            if sortedOrder and sortedOrder != 'None':
                foundIds.sort(
                    lambda ob1, ob2, attr = sortedOrder:
                    cmp(getattr(ob1, attr), getattr(ob2, attr)))
               
            
            return [o.getId() for o in foundIds]
        finally:
            self.pullContext()
    findObjectIds.useAutoDeserialize = 1


contactsServer = ContactsServer()


if __name__ == "__main__":
    contactsServer.launch(applicationName, applicationRole)
