# Glasnost
# By: Odile Bnassy <obenassy@entrouvert.com>
#     Thierry Dulieu <tdulieu@easter-eggs.com>
#     Frdric Pters <fpeters@theridion.com>
#     Benjamin Poussin <poussin@codelutin.com>
#     Emmanuel Raviart <eraviart@entrouvert.com>
#     Emmanuel Saracco <esaracco@easter-eggs.com>
#
# Copyright (C) 2000, 2001 Easter-eggs & Emmanuel Raviart
# Copyright (C) 2002 Odile Bnassy, Code Lutin, Thierry Dulieu, Easter-eggs,
#     Entr'ouvert, Frdric Pters, Benjamin Poussin, Emmanuel Raviart,
#     Emmanuel Saracco & Thridion
# Copyright (C) 2003 Odile Bnassy, Code Lutin, Thierry Dulieu, Easter-eggs,
#     Entr'ouvert, Ouvaton, Frdric Pters, Benjamin Poussin, Rodolphe
#     Quideville, Emmanuel Raviart, Emmanuel Saracco, Thridion & Vecam
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""Glasnost Authentication Proxy.

Classes:
    
    *AuthenticationProxy* -- The Authentication Proxy.

"""

__version__ = '$Revision: 1.13 $'[11:-2]


import glasnost.common.context as context

from DispatcherProxy import callServer, getApplicationToken
from ObjectsProxy import Proxy
from tools import *


class AuthenticationProxy(Proxy):
    
    """Wrap the Glasnost Authentication Proxy.

    This class is designed for authentication tests.
    In fact, it does not handle the authentication, it just provides methods in order to handle the authentication manualy.

    Methods:

    *canLogin*:
        Test the application login possibility.

    *computeServerIdFromUserToken*:
        Return the server ID.

    *getUserId*:
        Return the current user ID.
        
    *getUserIdAndToken*:
        Return the current user ID and Token.

    *setContainsUser*:
        Test if a user is in a user set.
    
    """
    
    serverRole = 'authentication'

    def canLogin(self, serverId = None):
        """Test if the application can log into the server.

        Keyword argument:
        
        *serverId*:
            The Server Id where to test log in (Default: None).
        
        Returns values:
        
        *true*:
            If the application can log in.
            
        *false*:
            If the application cannot log in.
            
        The application token is taked from the current context(s).
        
        """
        serverId = self.getServerId(serverId = serverId)
        return callServer(
            serverId,
            'canLogin',
            [serverId, getApplicationToken()])

    def computeServerIdFromUserToken(self, userToken):
        """Give back the ServerId from the UserToken.
        
        Keyword argument:
        
        *userToken*:
            A valid user token.
        
        Return values:
        
        The server ID string.
            
        The string always begin with 'glasnost://', with or without errors.
            
        """
        
        return 'glasnost://%s/%s' % (
            extractApplicationHostNameAndPort(userToken),
            self.serverRole)

    def getUserId(self):
        """Give back the current user ID from the top context.
        

        Return values:
        
        *''*:
            If the context don't contain any user Token.
            If the user token has expired
        
        Exceptions:

        *FaultUnknownUserToken*:
            The given user token doesn't exists (invalid).
                
        """
        
        userToken = context.getVar('userToken', default = '')
        if not userToken:
            return ''
        serverId = self.computeServerIdFromUserToken(userToken)
        return callServer(
            serverId,
            'getUserId',
            [getApplicationToken(), userToken])

    def getUserIdAndToken(self, userName, userPassword, serverId = None):
        """Give back the user ID and token from top context.
        
        Keywords arguments:

        *userName*:
            The Glasnost user name.

        *userPassword*:
            The Glasnost user password of userName.

        *serverId*:
            The Glasnost server Id to use (Default: None).

        Return values:

        *Array*:
            Containing the user ID and the user Token, in this order.
            
        Exceptions:

        **WRITEME**.
            
        """
        
        serverId = self.getServerId(serverId = serverId)
        return callServer(
            serverId,
            'getUserIdAndToken',
            [serverId, getApplicationToken(), utf8(userName),
             utf8(userPassword)])

    def setContainsUser(self, set, serverId = None):
        """Test if the set of users of an object contains the user.

        Keywords arguments:

        *set*:
            An valid object's set of users.

        *serverId*:
            The id of the server where to test (defaut: None).

        Return values:

        *0*: 
            If there is no user token in the context.
            If the set is empty (aka None)
            If the set doesn't contain the user.
                
        *1*:
            If the set contains the user.
            
        Exceptions:

        The same as the getUserId method.
        So, **WRITEME**.
        
        """
        userToken = context.getVar('userToken', default = '')
        if not userToken:
            return 0
        if set is None:
            return 0
        serverId = self.getServerId(serverId = serverId)
        return callServer(
            serverId,
            'setContainsUser',
            [serverId, getApplicationToken(), userToken, set])

