# Glasnost
# By: Odile Bnassy <obenassy@entrouvert.com>
#     Thierry Dulieu <tdulieu@easter-eggs.com>
#     Frdric Pters <fpeters@theridion.com>
#     Benjamin Poussin <poussin@codelutin.com>
#     Emmanuel Raviart <eraviart@entrouvert.com>
#     Emmanuel Saracco <esaracco@easter-eggs.com>
#
# Copyright (C) 2000, 2001 Easter-eggs & Emmanuel Raviart
# Copyright (C) 2002 Odile Bnassy, Code Lutin, Thierry Dulieu, Easter-eggs,
#     Entr'ouvert, Frdric Pters, Benjamin Poussin, Emmanuel Raviart,
#     Emmanuel Saracco & Thridion
# Copyright (C) 2003 Odile Bnassy, Code Lutin, Thierry Dulieu, Easter-eggs,
#     Entr'ouvert, Ouvaton, Frdric Pters, Benjamin Poussin, Rodolphe
#     Quideville, Emmanuel Raviart, Emmanuel Saracco, Thridion & Vecam
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


__doc__ = """Glasnost Dispatcher Proxy"""

__version__ = '$Revision: 1.28 $'[11:-2]


import sys
import xmlrpclib
import socket
import marshal
import os

import glasnost.common.faults as faults
import glasnost.common.context as context

from tools import *


mainModule = sys.modules['__main__']
serverAccessors = {}
serverClients = {}

try:
    socket.sslerror
except AttributeError:
    socket.sslerror = None


class FastServerProxy:
    def __init__(self, transport=None, encoding=None, verbose=0):
        self.__transport = transport
        self.__encoding = encoding
        self.__verbose = verbose

    def __request(self, methodname, params):
        request = {
            'methodName': methodname,
            'params':     params
        }
        request = marshal.dumps(request)

        response = self.__transport.request(
            self.__host,
            self.__handler,
            request,
            verbose=self.__verbose
            )

        if len(response) == 1:
            response = response[0]

        return response

    def __getattr__(self, name):
        # magic method dispatcher
        return xmlrpclib._Method(self.__request, name)


class FastTransport:
    """Handles a short-circuited connection to an XML-RPC server"""

    def __init__(self, host, port):
        self.address = '/tmp/.glasnost:%s' % port
    
    def request(self, host, handler, request_body, verbose=0):
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        s.connect(self.address)
        s.send('%s\n' % len(request_body))
        s.send(request_body)
        length = int(s.recv(12))
        r = ''
        while len(r) != length:
            r += s.recv(50000)
        r = marshal.loads(r)
        if type(r) is type([]):
            fault = faults.Fault(r[0], r[1])
            if fault.faultCode in faults.faults.keys():
                fault.__class__ = faults.faults[fault.faultCode]
            raise fault
        return r
    

class Lazy:
    resultHandler = None
    value = None

    def __call__(self):
        if self.resultHandler:
            return self.resultHandler(Lazy(self.value))
        else:
            if type(self.value) == type({}):
                # Value is a fault.
                fault = faults.Fault(self.value['faultCode'],
                                     self.value['faultString'])
                if fault.faultCode in faults.faults.keys():
                    fault.__class__ = faults.faults[fault.faultCode]
                raise fault
            else:
                # Value is a sequence containing the real value.
                return self.value[0]

    def __init__(self, value, resultHandler = None):
        self.value = value
        if resultHandler is not None:
            self.resultHandler = resultHandler


class MultiCall:
    calls = None
    earlyResults = None
    resultHandlers = None
    serverHostName = None
    serverPort = None

    def __init__(self):
        self.calls = []
        self.earlyResults = {}
        self.resultHandlers = []

    def add(self, serverAccessor, functionName, arguments, resultHandler):
        if self.serverHostName is None:
            self.serverHostName = serverAccessor['serverHostName']
            self.serverPort = serverAccessor['serverPort']
        else:
            assert self.serverHostName == serverAccessor['serverHostName']
            assert self.serverPort == serverAccessor['serverPort']
        self.calls.append({
            'methodName': functionName,
            'params': arguments,
            })
        self.resultHandlers.append(resultHandler)

    def addEarlyFault(self, earlyFault):
        i = len(self.calls) + len(self.earlyResults)
        self.earlyResults[i] = {
            'faultCode': earlyFault.faultCode,
            'faultString': earlyFault.faultString,
            }

    def addEarlyResult(self, earlyResult):
        i = len(self.calls) + len(self.earlyResults)
        self.earlyResults[i] = [earlyResult]

    def call(self):
        handledResults = []
        if self.calls:
            try:
                results = _callServer(self.serverHostName, self.serverPort,
                                      'system.multicall', [self.calls])
            except faults.Fault, fault:
                if fault.faultCode in faults.faults.keys():
                    fault.__class__ = faults.faults[fault.faultCode]
                raise
            for i in range(len(self.resultHandlers)):
                resultHandler = self.resultHandlers[i]
                result = results[i]
                handledResult = Lazy(result, resultHandler = resultHandler)
                handledResults.append(handledResult)
        earlyResultIndexes = self.earlyResults.keys()
        earlyResultIndexes.sort()
        for i in earlyResultIndexes:
            handledResults.insert(i, Lazy(self.earlyResults[i]))
        return handledResults


def _callServer(serverHostName, serverPort, functionName, arguments,
                mayTrySSL = 1):
    serverHostNameAndPort = '%s:%s' % (serverHostName, serverPort)
    if not serverClients.has_key(serverHostNameAndPort):
        if getConfig('Misc', 'BypassXmlRpc', default = 'true') == 'true' \
               and not isRemote(serverHostName):
            # should test capabilities here (?)
            transport = FastTransport(serverHostName, serverPort)
            if os.path.exists(transport.address):
                serverClients[serverHostNameAndPort] = FastServerProxy(
                    transport = transport)
        else:
            if mayTrySSL and socket.sslerror and \
                    (functionName == 'callGateway'
                     or functionName == 'getServerAccessor'):
                serverClients[serverHostNameAndPort] = xmlrpclib.ServerProxy(
                    'https://%s' % serverHostNameAndPort)
            else:
                serverClients[serverHostNameAndPort] = xmlrpclib.ServerProxy(
                    'http://%s' % serverHostNameAndPort)
    serverClient = serverClients[serverHostNameAndPort]
    try:
        result = apply(getattr(serverClient, functionName), arguments)
    except socket.sslerror:
        del serverClients[serverHostNameAndPort]
        return _callServer(
            serverHostName, serverPort, functionName, arguments, mayTrySSL = 0)
    return result


def callServer(serverId, functionName, arguments, resultHandler = None,
               multiCall = None):
    assert serverId is not None
    if serverId == context.getVar('applicationId'):
        server = context.getVar('server')
        result = server.xmlRpcServer.dispatch_call(
            functionName, arguments, isDirectCall = 1)
        if multiCall:
            multiCall.addEarlyResult(result)
            return None
        return result
    if not serverAccessors.has_key(serverId):
        serverAccessors[serverId] = getServerAccessor(serverId)
    serverAccessor = serverAccessors[serverId]
    if serverAccessor.has_key('isGateway') and serverAccessor['isGateway']:
        arguments = [serverId, functionName, arguments]
        functionName = 'callGateway'
    if multiCall is not None:
        # Instead of calling directly, pack all the calls in a multiCall.
        multiCall.add(serverAccessor, functionName, arguments, resultHandler)
        return None
    if resultHandler is None:
        try:
            return _callServer(
                serverAccessor['serverHostName'], serverAccessor['serverPort'],
                functionName, arguments)
        except faults.Fault, fault:
            if fault.faultCode in faults.faults.keys():
                fault.__class__ = faults.faults[fault.faultCode]
            raise
    else:
        try:
            result = [_callServer(
                serverAccessor['serverHostName'], serverAccessor['serverPort'],
                functionName, arguments)]
        except faults.Fault, fault:
            result = {
                'faultCode': fault.faultCode,
                'faultString': fault.faultString,
                }
        result = resultHandler(Lazy(result))
        return result
    

def getApplicationId(applicationToken):
    return callServer(
        getDefaultDispatcherId(),
        'getApplicationId',
        [getApplicationToken(), applicationToken])


def getApplicationToken():
    applicationToken = context.getVar('applicationToken')
    if applicationToken is None:
        applicationId = context.getVar('applicationId')
        dispatcherId = context.getVar('dispatcherId')
        applicationToken = callServer(
            dispatcherId,
            'getApplicationToken',
            [applicationId])
        context.setVar('applicationToken', applicationToken)
    return applicationToken


def getServerAccessor(serverId):
    serverHostNameAndPort = extractApplicationHostNameAndPort(serverId)
    serverInfos = serverHostNameAndPort.split(':', 1)
    serverHostName = serverInfos[0]
    if serverHostName == 'system':
        serverHostName = getConfig('Dispatcher', 'ServerHostName')
    if len(serverInfos) < 2:
        serverPort = eval(getConfig('Dispatcher', 'ServerPort'), {})
    else:
        serverPort = int(serverInfos[1])
    try:
        return _callServer(serverHostName, serverPort, 'getServerAccessor',
                           [serverId])
    except faults.Fault, fault:
        if fault.faultCode in faults.faults.keys():
            fault.__class__ = faults.faults[fault.faultCode]
        raise


def getRegisteredRoles(dispatcherId):
    return callServer(
        dispatcherId,
        'getRegisteredRoles',
        [dispatcherId])


def isRemote(hostName):
    if hostName in (socket.gethostname(), socket.getfqdn()):
        return 0
    knownDispatchers = getConfig('Misc', 'DispatcherURLs', '').split(',')
    if hostName in [x.strip()[11:].split('/')[0]
                    for x in knownDispatchers] + ['localhost']:
        return 0
    return hostName not in [
        x.strip()
        for x in getConfig('Misc', 'LocalHostNames', '').split(',')]


def registerServer(serverHostName, serverPort):
    dispatcherId = context.getVar('dispatcherId')
    callServer(
        dispatcherId,
        'registerServer',
        [getApplicationToken(), serverHostName, serverPort])


def unregisterServer():
    dispatcherId = context.getVar('dispatcherId')
    callServer(
        dispatcherId,
        'unregisterServer',
        [getApplicationToken()])
