# Glasnost
# By: Odile Bnassy <obenassy@entrouvert.com>
#     Thierry Dulieu <tdulieu@easter-eggs.com>
#     Frdric Pters <fpeters@theridion.com>
#     Benjamin Poussin <poussin@codelutin.com>
#     Emmanuel Raviart <eraviart@entrouvert.com>
#     Emmanuel Saracco <esaracco@easter-eggs.com>
#
# Copyright (C) 2000, 2001 Easter-eggs & Emmanuel Raviart
# Copyright (C) 2002 Odile Bnassy, Code Lutin, Thierry Dulieu, Easter-eggs,
#     Entr'ouvert, Frdric Pters, Benjamin Poussin, Emmanuel Raviart,
#     Emmanuel Saracco & Thridion
# Copyright (C) 2003 Odile Bnassy, Code Lutin, Thierry Dulieu, Easter-eggs,
#     Entr'ouvert, Ouvaton, Frdric Pters, Benjamin Poussin, Rodolphe
#     Quideville, Emmanuel Raviart, Emmanuel Saracco, Thridion & Vecam
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


"""Glasnost People Proxy.

Classes:
    
*AdminPeople*:
     People proxy administrative class.

*Person*:
    People proxy object.

*PeopleProxy*:
    People server Proxy.

"""

__version__ = '$Revision: 1.17 $'[11:-2]


from glasnost.common.PeopleCommon import *
import glasnost.common.context as context

from DispatcherProxy import callServer, getApplicationToken
from ObjectsProxy import register, AdminMixin, ObjectProxyMixin, ObjectsProxy
from tools import *


class AdminPeople(AdminMixin, AdminPeopleCommon):
    
    """Management of the people proxy administrative informations."""
    
    def getProxy(self):
        """Return the People Proxy class instance."""
        
        return peopleProxy
        
register(AdminPeople)


class Person(ObjectProxyMixin, PersonCommon):

    """The People Proxy's object class."""

    def getProxy(self):
        """Return the People Proxy class instance."""
        
        return peopleProxy
        
register(Person)


class PeopleProxy(PeopleCommon, ObjectsProxy):

    """Proxy of the People server.

    This class is used in order to communicate with the server.
    
    Methods:
    ========

    *abstainForVote*:
        Vote as an abstaint.

    *canEmailPassword*:
        Test password is emailable.

    *changePassword*:
        Change the people password.

    *checkObjectAuthentication*:
        Check the people object login/password.

    *computeServerIdFromUserToken*:
        Generate the server ID string.
        
    *emailPassword*:
        Email the people password.

    *findObjectIds*:
        Search people objects with a string.

    *getObjectByLogin*:
        Get the people object from login.

    *hasObjectByLogin*:
        Test if an people object uses a login.

    *setLoginTime*:
        Timestamp the login.
        
    *vote*:
        Vote.
        
    Attributes:
    ==========
    
    *adminClassName*:
        The Administrator wrapper people class name.
        
    *newObjectNameCapitalized*:
        The gettext string illustrating a new user.
    
    *objectClassName*:
        The normal people wrapper class name.
    
    *objectName*:
        The gettext string of the people object name.
    
    *objectNameCapitalized*:
        The gettext string of the people object name, capitalzed.

    *objectsName*:
        The gettext string of the collection of people objects.
    
    *objectsNameCapitalized*:
        The gettext string of the collection of people objects, capitalized.
    
    *serverRole*:
        The name of the server.
    
    """

    def abstainForVote(self, electionId):
        """Abstain for vote for election election ID.
        
        Keywords arguments:

        *electionId*:
            The ID of the election that the people abstains.

        The user token is taken from the context.
        Nothing appends if the election ID is invalid.
        
        """
        
        userToken = context.getVar('userToken', default = '')
        serverId = self.computeServerIdFromUserToken(userToken)
        callServer(
            serverId,
            'abstainForVote',
            [serverId, getApplicationToken(), userToken, electionId])

    def canEmailPassword(self, objectId):
        """Test if the object is able email the people password.
        
        Keyword arguments:

        *objectId*:
            The object ID refering to a valid object (with login and email).

        Return values:
        
        *0*:
            The object cannot send the people password by email. (Generally because it have nor email nor login attributes; cf PeopleServer.py).
            
        *1*:
            The object can send the people password by email.
            
        Exceptions:

        **WRITEME**.
            
        """
        
        userToken = context.getVar('userToken', default = '')
        return callServer(
            extractServerId(objectId),
            'canEmailPassword',
            [extractServerId(objectId), getApplicationToken(), userToken,
             objectId])

    def changePassword(self, objectId):
        """Change the object (user) password.
        
        Generate a new password and send it by e-mail.
        
        Keyword argument:

        *objectID*:
            The ID of the people object that will 
            have a new generated password.
            Of course, this object must have a login.
            And a valid e-mail.
            

        Return values:
        
        The new password.
            
        Exceptions:

        *FaultMissingItem*:
            The object doesn't exist.

        *FaultUserAccessDenied*:
            The user is not in the set of authorized people for password change.
        
        """
        
        userToken = context.getVar('userToken', default = '')
        return iso8859_15(callServer(
            extractServerId(objectId),
            'changePassword',
            [extractServerId(objectId), getApplicationToken(), userToken,
             objectId]))

    def checkObjectAuthentication(self, objectLogin, objectPassword,
                                  serverId = None):
        """Return the object ID of an object login/password pair.

        Keywords arguments:

        *objectLogin*:
            The object login.

        *objectPassword*:
            The object password.

        Return value:
        
            The object ID.

        Exceptions:
        
        *FaultWrongLogin*:
            If the object login is invalid (doesn't exist).

        *FaultWrongPassword*:
            If the object password doesn't agree with object login.
        
        """
        
        serverId = self.getServerId(serverId = serverId)
        return callServer(
            serverId, 
            'checkObjectAuthentication',
            [serverId, getApplicationToken(), utf8(objectLogin),
             utf8(objectPassword)])

    def computeServerIdFromUserToken(self, userToken):
        """Return the user's server ID string from his token."""
        
        return 'glasnost://%s/%s' % (
            extractApplicationHostNameAndPort(userToken),
            self.serverRole)

    def emailPassword(self, objectId):
        """Send the object (user) password by email.
        
        Keyword argument:
            
        *objectId*:
            The id of the object.

        Exceptions:

        *FaultMissingItem*:
            The object with ID objectId does not exist.

        *FaultBadEmailAdress*:
            The object does not contain login or email adresse.
        
        """
        
        userToken = context.getVar('userToken', default = '')
        callServer(
            extractServerId(objectId),
            'emailPassword',
            [extractServerId(objectId), getApplicationToken(), userToken,
             objectId])

    def findObjectIds(self, text, serverId = None):
        """Search objects (people) in plain text.

        Find people objects by calling the server function wich compare the texts attributes of the objects with the text string.
        The user token is taken from the context(s).

        Keyword arguments:

        *text*:
            The search string.

        *serverId*:
            The server ID (default: None).

        Return values:

        *empty array*:
            If the search string is empty.
            If no objects was found.

        *objects ID array*:
            If objects were found.

        Exceptions:

        **WRITEME**.
        
        """
        
        userToken = context.getVar('userToken', default = '')
        serverId = self.getServerId(serverId = serverId)
        if not text:
            return []
        return callServer(
            serverId,
            'findObjectIds',
            [serverId, getApplicationToken(), userToken, utf8(text)])

    def getObjectByLogin(self, objectLogin, serverId = None):
        """Get the people object instance from the people login.
        
        Keyword arguments:
        
        *objectLogin*:
            The user login.

        *serverId*:
            The server ID.

        Return value:

        The people object instance corresponding to the given login.
        
        Exceptions:

        *FaultMissingItem*:
            There is no object corresponding to the given object login.
        
        """
        
        userToken = context.getVar('userToken', default = '')
        serverId = self.getServerId(serverId = serverId)
        objectImport = callServer(
            serverId,
            'getObjectByLogin',
            [serverId, getApplicationToken(), userToken, utf8(objectLogin)])
        object = self.newThing('object', objectImport['__thingName__'])
        object.importFromXmlRpc(objectImport)
        return object

    def hasObjectByLogin(self, objectLogin, serverId = None):
        """Test if there is a people object with a specific login.
        
        Keywords arguments:

        *objectLogin*:
            The people object login.

        *serverId*:
            The server ID (default: None).

        Return values:

        *0*:
            No object with login == objectLogin was found.

        *1*:
            There is a people object with the given login.
            
        """
        
        userToken = context.getVar('userToken', default = '')
        serverId = self.getServerId(serverId = serverId)
        return callServer(
            serverId,
            'hasObjectByLogin',
            [serverId, getApplicationToken(), userToken, utf8(objectLogin)])

    def setLoginTime(self, serverId = None):
        """Set the login timestamp.

        This timestamp applies on the user designed by the user token saved in the context.

        Keyword argument:

        *serverId*:
            The server ID (default=None).

        Exceptions:

        *Fault*:
            Any fault except if the user token is unknown.
            In this last case, no exception is raised.
            But of course, it doesn't set the login time :-).
        
        *FaultMissingItem*:
            The user ID associated with the user token does not exist.
        
        """
        
        userToken = context.getVar('userToken', default = '')
        serverId = self.getServerId(serverId = serverId)
        callServer(
            serverId,
            'setLoginTime',
            [serverId, getApplicationToken(), userToken])

    def vote(self, electionId, voteToken):
        """Vote.
        
        Keywords arguments:

        *electionId*:
            The ID of the election where to vote.

        *voteToken*:
            The token wich allow to vote.

        Return value:
        
        *0*:
            Everytime.

        Exception:

        *FaultMissingItem*:
            The voter ID associated with the voter token is not valid. (Hey, Dave, What are you doing ?)
        
        """
        
        userToken = context.getVar('userToken', default = '')
        serverId = self.computeServerIdFromUserToken(userToken)
        callServer(
            serverId,
            'vote',
            [serverId, getApplicationToken(), userToken, electionId,
             voteToken])

