# Glasnost
# By: Odile Bnassy <obenassy@entrouvert.com>
#     Thierry Dulieu <tdulieu@easter-eggs.com>
#     Frdric Pters <fpeters@theridion.com>
#     Benjamin Poussin <poussin@codelutin.com>
#     Emmanuel Raviart <eraviart@entrouvert.com>
#     Emmanuel Saracco <esaracco@easter-eggs.com>
#
# Copyright (C) 2000, 2001 Easter-eggs & Emmanuel Raviart
# Copyright (C) 2002 Odile Bnassy, Code Lutin, Thierry Dulieu, Easter-eggs,
#     Entr'ouvert, Frdric Pters, Benjamin Poussin, Emmanuel Raviart,
#     Emmanuel Saracco & Thridion
# Copyright (C) 2003 Odile Bnassy, Code Lutin, Thierry Dulieu, Easter-eggs,
#     Entr'ouvert, Ouvaton, Frdric Pters, Benjamin Poussin, Rodolphe
#     Quideville, Emmanuel Raviart, Emmanuel Saracco, Thridion & Vecam
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


__doc__ = """Glasnost Objects Server"""

__version__ = '$Revision: 1.85 $'[11:-2]


import copy
import fcntl
import getopt
import os
import shutil
import signal
import sys
from threading import Lock
import traceback

import glasnost.common.context as context
import glasnost.common.faults as faults
from glasnost.common.ObjectsCommon import AdminCommon
import glasnost.common.tools_new as commonTools

import kinds # Do not remove!
import things
from tools import *
from xmlrpcServer import xmlrpcServer, xmlrpcThreadingServer, xmlrpcFastServer, xmlrpcFastThreadingServer

# Shall be located before the proxy imports.
context.clear()

from glasnost.proxy.CacheProxy import invalidateValue
from glasnost.proxy.DispatcherProxy import getApplicationId, getApplicationToken, registerServer
from glasnost.proxy.GroupsProxy import getSetContainedIds


### FIXME: To remove ASAP.
mainModule = sys.modules['__main__']


register = things.register


class ObjectServerMixin(things.ThingMixin):
    def __getinitargs__(self):
        # Ensure that __init__ is called during unpickling.
        return ()

    def acquireNonCore(self, objectDirectoryPath = None,
                       dataDirectoryPath = None, parentSlot = None):
        if objectDirectoryPath is None:
            server = self.getServer()
            virtualServerId = server.computeVirtualServerId(self.id)
            virtualServer = server.getVirtualServer(virtualServerId)
            if dataDirectoryPath is None:
                dataDirectoryPath = virtualServer.dataDirectoryPath
            objectsDirectoryPath = os.path.join(
                dataDirectoryPath, server.applicationRole)
            serverHostNameAndPort, serverRole, localId = splitId(
                self.id)
            objectDirectoryPath = os.path.join(
                objectsDirectoryPath, localId)
        things.ThingMixin.acquireNonCore(
            self, objectDirectoryPath, parentSlot = parentSlot)

    def canBeCreatedByClient(self, clientId):
        clientNameAndPort, clientRole = splitApplicationId(clientId)
        return clientRole == 'workflows'

    def canBeDeletedByClient(self, clientId):
        return 0
    
    def canBeGottenByClient(self, clientId):
        clientNameAndPort, clientRole = splitApplicationId(clientId)
        return clientRole == 'workflows'

    def canBeModified(self):
        return 1

    def canBeModifiedByClient(self, clientId):
        clientNameAndPort, clientRole = splitApplicationId(clientId)
        return clientRole == 'workflows'

    def checkAddIsPossible(self):
        pass

    def checkModifyIsPossible(self, changes, givenSlotNames = None):
        if self.version != changes.version:
            raise faults.WrongVersion()

    def getServer(self):
        return context.getVar('server')

    def initPartial(self, partial, givenSlotNames = None):
        pass

    def modify(self, changes, givenSlotNames = None):
        self.version += 1
        for slotName in self.getModifySlotNames():
            slot = self.getSlot(slotName)
            kind = slot.getKind()
            if kind.isAutomaticalyModified:
                kind.setAutomaticalValue(slot)
                continue
            if not kind.hasToModify:
                continue
            if givenSlotNames and not slotName in givenSlotNames:
                continue
            changesSlot = changes.getSlot(slotName)
            slot.setValue(changesSlot.getValue())

    def removeNonCore(self, objectDirectoryPath = None,
                      dataDirectoryPath = None, parentSlot = None):
        if objectDirectoryPath is None:
            server = self.getServer()
            virtualServerId = server.computeVirtualServerId(self.id)
            virtualServer = server.getVirtualServer(virtualServerId)
            if dataDirectoryPath is None:
                dataDirectoryPath = virtualServer.dataDirectoryPath
            if not os.access(dataDirectoryPath, os.F_OK):
                return
            objectsDirectoryPath = os.path.join(
                dataDirectoryPath, server.applicationRole)
            if not os.access(objectsDirectoryPath, os.F_OK):
                return
            serverHostNameAndPort, serverRole, localId = splitId(self.id)
            objectDirectoryPath = os.path.join(objectsDirectoryPath, localId)
        if not os.access(objectDirectoryPath, os.F_OK):
            return
        for slotName in self.getSlotNames(parentSlot = parentSlot):
            slot = self.getSlot(slotName, parentSlot = parentSlot)
            kind = slot.getKind()
            if kind.useFileStorage:
                shutil.rmtree(objectDirectoryPath)

    def saveNonCore(self, objectDirectoryPath = None, dataDirectoryPath = None,
                    parentSlot = None):
        if objectDirectoryPath is None:
            server = self.getServer()
            virtualServerId = server.computeVirtualServerId(self.id)
            virtualServer = server.getVirtualServer(virtualServerId)
            if dataDirectoryPath is None:
                dataDirectoryPath = virtualServer.dataDirectoryPath
            if not os.access(dataDirectoryPath, os.F_OK):
                os.mkdir(dataDirectoryPath)
                os.chmod(dataDirectoryPath, 0750)
            objectsDirectoryPath = os.path.join(
                dataDirectoryPath, server.applicationRole)
            if not os.access(objectsDirectoryPath, os.F_OK):
                os.mkdir(objectsDirectoryPath)
                os.chmod(objectsDirectoryPath, 0750)
            serverHostNameAndPort, serverRole, localId = splitId(
                self.id)
            objectDirectoryPath = os.path.join(
                objectsDirectoryPath, localId)
            if os.access(objectDirectoryPath, os.F_OK):
                try:
                    shutil.rmtree(objectDirectoryPath)
                except OSError:
                    # FIXME: If objectDirectoryPath is not a directory, do
                    # nothing.
                    # Happens for articles, books, uploadFiles... which
                    # should be converted to the new way of handling 
                    print 'Not a directory: %s' % objectDirectoryPath
        things.ThingMixin.saveNonCore(
            self, objectDirectoryPath, parentSlot = parentSlot)

    def setAutomaticalSlots(self, parentSlot = None):
        for slotName in self.getModifySlotNames():
            slot = self.getSlot(slotName)
            kind = slot.getKind()
            if kind.isAutomaticalyModified:
                kind.setAutomaticalValue(slot, init = 1)


class AdminServerWithoutWritersMixin(ObjectServerMixin):
    def newDefaultObject(self):
        server = self.getServer()
        return server.newThing('object', server.objectClassName)


class AdminServerMixin(AdminServerWithoutWritersMixin):
    pass


class VirtualServer:
    dataDirectoryPath = None
    virtualServerId = None
    lock = None

    def __getinitargs__(self):
        # Ensure that __init__ is called during unpickling.
        return ()

    def __getstate__(self):
        # Don't pickle the threading locks.
        state = self.__dict__.copy()
        if state.has_key('lock'):
            del state['lock']
        if state.has_key('dataDirectoryPath'):
            del state['dataDirectoryPath']
        return state

    def __init__(self):
        self.lock = Lock()

    def check(self, server, virtualServerId):
        pass
        #assert self.virtualServerId == virtualServerId

    def convertIds(self, sourceDispatcherId, destinationDispatcherId):
        if self.virtualServerId is not None:
            sourceHostName = extractApplicationHostName(sourceDispatcherId)
            destinationHostName = extractApplicationHostName(
                destinationDispatcherId)
            if self.dataDirectoryPath is not None \
               and self.dataDirectoryPath.endswith(sourceHostName):
                self.dataDirectoryPath = self.dataDirectoryPath[
                    :-len(sourceHostName)] + destinationHostName
            newVirtualServerId = self.virtualServerId.replace(
                sourceDispatcherId, destinationDispatcherId)
            if newVirtualServerId != self.virtualServerId:
                self.virtualServerId = newVirtualServerId

    def exportNonCore(self, exportDirectoryPath):
        pass

    def importNonCore(self, importDirectoryPath):
        pass

    def init(self, server, virtualServerId):
        self.virtualServerId = virtualServerId

    def initFromOldData(self, server, virtualServerId, data):
        self.virtualServerId = virtualServerId

    def removeIds(self, rolesToKeep):
        pass


class ObjectsVirtualServer(VirtualServer):
    admin = None
    nextLocalId = 1
    objects = None

    def convertIds(self, sourceDispatcherId, destinationDispatcherId):
        VirtualServer.convertIds(
            self, sourceDispatcherId, destinationDispatcherId)
        self.admin.convertIds(sourceDispatcherId, destinationDispatcherId)
        for id, object in self.objects.items():
            newId = id.replace(sourceDispatcherId, destinationDispatcherId)
            if newId != id:
                del self.objects[id]
                self.objects[newId] = object
            object.convertIds(sourceDispatcherId, destinationDispatcherId)
            sourceHostName = extractApplicationHostName(sourceDispatcherId)
            destinationHostName = extractApplicationHostName(
                destinationDispatcherId)
            if self.dataDirectoryPath is not None \
               and self.dataDirectoryPath.endswith(destinationHostName):
                sourceDataDirectoryPath = self.dataDirectoryPath[
                    :-len(destinationHostName)] + sourceHostName
            else:
                sourceDataDirectoryPath = self.dataDirectoryPath
            object.acquireNonCore(dataDirectoryPath = sourceDataDirectoryPath)
            object.convertIds(sourceDispatcherId, destinationDispatcherId)
            object.saveNonCore()
            object.releaseNonCore()

    def exportNonCore(self, exportDirectoryPath):
        VirtualServer.exportNonCore(self, exportDirectoryPath)
        for object in self.objects.values():
            object.acquireNonCore()
            object.saveNonCore(dataDirectoryPath = exportDirectoryPath)
            object.releaseNonCore()

    def getNextLocalId(self):
        result = self.nextLocalId
        self.nextLocalId += 1
        return result

    def importNonCore(self, importDirectoryPath):
        VirtualServer.importNonCore(self, importDirectoryPath)
        for object in self.objects.values():
            object.acquireNonCore(dataDirectoryPath = importDirectoryPath)
            object.saveNonCore()
            object.releaseNonCore()

    def init(self, server, virtualServerId):
        VirtualServer.init(self, server, virtualServerId)
        self.admin = server.newAdmin()
        if virtualServerId:
            self.admin.id = '%s/__admin__' % virtualServerId
        else:
            self.admin.id = '%s/%s/__admin__' % (
                server.dispatcherIds[0], server.applicationRole)
        self.objects = {}

    def initFromOldData(self, server, virtualServerId, data):
        VirtualServer.initFromOldData(self, server, virtualServerId, data)
        (self.nextLocalId,
         self.objects,
         self.admin,
         ) = data

    def removeIds(self, rolesToKeep):
        VirtualServer.removeIds(self, rolesToKeep)
        self.admin.removeIds(rolesToKeep)
        for object in self.objects.values():
            object.removeIds(rolesToKeep)


class Server(things.BaseThing):
    VirtualServer = None
    applicationName = None
    applicationRole = None
    dataDirectoryPath = None
    dispatcherIds = None
    hostName = None
    isSubServer = 0
    isThreaded = 1
    isVerbose = 0
    port = None
    thingCategory = 'server'
    useDataFile = 1
    version = None
    virtualServers = None
    xmlRpcServer = None

    def addDispatcherXmlRpc(self, virtualServerId, clientToken,
                            newDispatcherId):
        # TODO: check on clientToken
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            newVirtualServerId = '%s/%s' % (
                newDispatcherId, self.applicationRole)
            self.virtualServers[newVirtualServerId] = self.loadVirtualServer(
                    newVirtualServerId)
        finally:
            self.pullContext()

    def canGetAdmin(self, virtualServerId, clientToken, userToken):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            return 0
        finally:
            self.pullContext()

    def computeVirtualServerId(self, objectId):
        return 'glasnost://%s/%s' % (
                extractApplicationHostNameAndPort(objectId),
                self.applicationRole)

    def convertVirtualServersIds(
        self, sourceDispatcherId, destinationDispatcherId):
        sourceVirtualServerId = '%s/%s' % (
            sourceDispatcherId, self.applicationRole)
        destinationVirtualServerId = '%s/%s' % (
            destinationDispatcherId, self.applicationRole)
        if not self.virtualServers.has_key(sourceVirtualServerId):
            print 'Missing virtual server for %s' % sourceDispatcherId
            return 2
        context.push()
        destinationVirtualServer = self.virtualServers[
            sourceVirtualServerId].__class__()
        for key, value in self.virtualServers[
                sourceVirtualServerId].__dict__.items():
            if not destinationVirtualServer.__dict__.has_key(key):
                destinationVirtualServer.__dict__[key] = value
        self.virtualServers[destinationVirtualServerId
                            ] = destinationVirtualServer
        context.setVar('server', self)
        destinationVirtualServer.convertIds(
            sourceDispatcherId, destinationDispatcherId)
        del self.virtualServers[sourceVirtualServerId]
        self.saveVirtualServer(destinationVirtualServer)
        return None

    def exportVirtualServer(self, virtualServerId, exportDirectoryPath):
        if not self.useDataFile:
            return None
        if virtualServerId is None:
            return None
        import gnosis.xml.pickle as glasnostPickle
        virtualServer = self.getVirtualServer(virtualServerId)
        if not os.access(exportDirectoryPath, os.F_OK):
            os.mkdir(exportDirectoryPath)
            os.chmod(exportDirectoryPath, 0750)
        xmlFilePath = os.path.join(
            exportDirectoryPath, self.applicationName + '.xml')
        xmlFile = open(xmlFilePath, 'wb')
        os.chmod(xmlFile, 0640)
        fcntl.lockf(xmlFile, fcntl.LOCK_EX)
        xmlFile.write(self.version + '\n')
        virtualServer.lock.acquire()
        glasnostPickle.dump(virtualServer, xmlFile, 0)
        virtualServer.lock.release()
        fcntl.lockf(xmlFile, fcntl.LOCK_UN)
        xmlFile.close()
        virtualServer.exportNonCore(exportDirectoryPath)
        return None

    def getVirtualServer(self, virtualServerId):
        return self.virtualServers[virtualServerId]

    def handleGetopt(self):
        try:
            opts, args = getopt.getopt(
                sys.argv[1:],
                'd:hi:s:x:y:v',
                ['destination=', 'export=', 'help', 'import=', 'source=',
                 'system=', 'verbose'])
        except getopt.GetoptError:
            self.usage()
            return 2
        if len(args) != 0:
            self.usage()
            return 2
        destinationDispatcherId = None
        doExport = 0
        doImport = 0
        doSystem = 0
        exportDirectoryPath = None
        importDirectoryPath = None
        sourceDispatcherId = None
        for o, a in opts:
            if o in ('-d', '--destination'):
                destinationDispatcherId = a
                if destinationDispatcherId[-1] == '/':
                    destinationDispatcherId = destinationDispatcherId[:-1]
                if not destinationDispatcherId.startswith('glasnost://') \
                   or '/' in destinationDispatcherId[len('glasnost://'):]:
                    print 'Wrong destination = %s' % destinationDispatcherId
                    return 2
            if o in ('-h', '--help'):
                usage()
                return 0
            if o in ( '-v', '--verbose'):
                self.isVerbose = 1
            if o in ('-s', '--source'):
                sourceDispatcherId = a
                if sourceDispatcherId[-1] == '/':
                    sourceDispatcherId = sourceDispatcherId[:-1]
                if not sourceDispatcherId.startswith('glasnost://') \
                   or '/' in sourceDispatcherId[len('glasnost://'):]:
                    print 'Wrong source = %s' % sourceDispatcherId
                    return 2
            if o in ('-i', '--import'):
                importDirectoryPath = a
                doImport = 1
            if o in ('-x', '--export'):
                exportDirectoryPath = a
                doExport = 1
            if o in ('-y', '--system'):
                doSystem = 1
                systemRoles = eval(a)
                if type(systemRoles) not in [type(()), type([])]:
                    print 'Wrong system roles = %s' % systemRoles
                    return 2
        if doExport:
            if not sourceDispatcherId:
                print 'Missing source to export'
                return 2
            if not self.useDataFile:
                return 0
            self.initOneDispatcher(sourceDispatcherId)
            self.load()
            virtualServerId = '%s/%s' % (
                sourceDispatcherId, self.applicationRole)
            hostName = extractApplicationHostName(virtualServerId)
            exitCode = self.exportVirtualServer(
                virtualServerId, os.path.join(exportDirectoryPath, hostName))
            if exitCode is not None:
                return exitCode
            return 0
        if doImport:
            if not destinationDispatcherId:
                print 'Missing destination to import into'
                return 2
            if not self.useDataFile:
                return 0
            self.initOneDispatcher(destinationDispatcherId)
            virtualServerId = '%s/%s' % (
                destinationDispatcherId, self.applicationRole)
            hostName = extractApplicationHostName(virtualServerId)
            virtualServer = self.importVirtualServer(
                virtualServerId, os.path.join(importDirectoryPath, hostName))
            if virtualServer is not None:
                self.virtualServers = {
                    virtualServerId: virtualServer,
                    }
                self.saveVirtualServer(virtualServer)
            return 0
        if doSystem:
            destinationDispatcherId = 'glasnost://system'
        if sourceDispatcherId and destinationDispatcherId:
            # Convert all the ids from sourceDispatcherId to
            # destinationDispatcherId.
            # If doSystem is set, also perform other modifications.
            if not self.useDataFile:
                return 0
            if doSystem and not self.applicationRole in systemRoles:
                return 0
            self.initOneDispatcher(sourceDispatcherId)
            self.load()
            exitCode = self.convertVirtualServersIds(
                sourceDispatcherId, destinationDispatcherId)
            if exitCode is not None:
                return exitCode
            if doSystem:
                exitCode = self.removeVirtualServerHistory(
                    destinationDispatcherId)
                if exitCode is not None:
                    return exitCode
                exitCode = self.removeIdsFromVirtualServer(
                    destinationDispatcherId, systemRoles)
                if exitCode is not None:
                    return exitCode
            return 0
        return None

    def handleSignal(self, signalNumber, stackFrame):
        if signalNumber == signal.SIGINT:
            raise KeyboardInterrupt
        if signalNumber == signal.SIGHUP:
            self.setLogFile()
            return
        if signalNumber == signal.SIGTERM:
            print 'Received signal SIGTERM: exiting.'
            os._exit(0)

    def importVirtualServer(self, virtualServerId, importDirectoryPath):
        if self.useDataFile:
            if virtualServerId is None:
                return None
            import gnosis.xml.pickle as glasnostPickle
            glasnostPickle.setParanoia(0)
            pickleFilePath = os.path.join(
                importDirectoryPath, self.applicationName + '.pickle')
            xmlFilePath = os.path.join(
                importDirectoryPath, self.applicationName + '.xml')
            rcFile = None
            if os.access(xmlFilePath, os.F_OK):
                rcFile = open(xmlFilePath, 'rb')
                import gnosis.xml.pickle as glasnostPickle
                glasnostPickle.setParanoia(0)
            elif os.access(pickleFilePath, os.F_OK):
                import cPickle as glasnostPickle
                rcFile = open(pickleFilePath, 'rb')
            if rcFile is None:
                virtualServer = self.VirtualServer()
                virtualServer.init(self, virtualServerId)
            else:
                fcntl.lockf(rcFile, fcntl.LOCK_SH)
                try:
                    versionLine = rcFile.readline(20).strip()
                    splittedVersion = versionLine.split('.')
                    version = 0.0
                    factor = 1000000.0
                    for number in splittedVersion:
                        version += int(number) * factor
                        factor = factor / 1000
                except ValueError:
                    version = 0
                    rcFile.seek(0)
                print 'file:',  rcFile
                data = glasnostPickle.load(rcFile)
                fcntl.lockf(rcFile, fcntl.LOCK_UN)
                rcFile.close()
                if isinstance(data, VirtualServer):
                    virtualServer = data
                else:
                    virtualServer = self.VirtualServer()
                    virtualServer.initFromOldData(self, virtualServerId, data)
                if version < 5004:
                    if not hasattr(virtualServer, 'virtualServerId') \
                       or not virtualServer.virtualServerId:
                        virtualServer.virtualServerId = \
                                                    virtualServer.applicationId
                        del virtualServer.applicationId
                virtualServer.check(self, virtualServerId)
                self.repairVirtualServer(virtualServer, version)
            hostName = extractApplicationHostName(virtualServerId)
            virtualServerDataDirectoryPath = getConfig(
                hostName, 'DataDirectoryPath')
            if not virtualServerDataDirectoryPath:
                virtualServerDataDirectoryPath = getConfig(
                    'Misc', 'DispatcherDataDirectoryPath',
                    vars = {'hostname': hostName})
            virtualServer.dataDirectoryPath = virtualServerDataDirectoryPath
        else:
            virtualServer = self.VirtualServer()
            virtualServer.init(self, virtualServerId)
        virtualServer.importNonCore(importDirectoryPath)
        return virtualServer

    def init(self):
        if self.useDataFile:
            self.dataDirectoryPath = getConfig('Misc', 'DataDirectoryPath')
        self.setDispatcherIds()

    def initOneDispatcher(self, dispatcherId):
        if self.useDataFile:
            self.dataDirectoryPath = getConfig('Misc', 'DataDirectoryPath')
        self.dispatcherIds = [dispatcherId]

    def launch(self, applicationName, applicationRole, hasToRun = 1):
        self.applicationName = applicationName
        self.applicationRole = applicationRole
        if not self.isSubServer:
            exitCode = self.handleGetopt()
            if exitCode is not None:
                sys.exit(exitCode)
            try:
                self.port = eval(
                    getConfig(self.applicationName, 'ServerPort'), {})
            except TypeError:
                if self.isVerbose:
                    sys.stderr.write(
                        'No server port set.  Not launching server (%s).\n'
                        % applicationName)
                sys.exit(1)
            goDaemon()
            self.setLogFile()
        self.init()
        self.load()
        if not self.isSubServer:
            self.startXmlRpcServer()
            if hasToRun:
                self.run()

    def load(self):
        self.virtualServers = {}
        for dispatcherId in self.dispatcherIds:
            virtualServerId = '%s/%s' % (dispatcherId, self.applicationRole)
            self.virtualServers[virtualServerId] = self.loadVirtualServer(
                virtualServerId)

    def loadVirtualServer(self, virtualServerId):
        if self.useDataFile:
            if virtualServerId is None:
                virtualServerDataDirectoryPath = self.dataDirectoryPath
            else:
                hostName = extractApplicationHostName(virtualServerId)
                virtualServerDataDirectoryPath = getConfig(
                    hostName, 'DataDirectoryPath')
                if not virtualServerDataDirectoryPath:
                    virtualServerDataDirectoryPath = getConfig(
                        'Misc', 'DispatcherDataDirectoryPath',
                        vars = {'hostname': hostName},
                        default = '/var/lib/glasnost/%s' % hostName)

            pickleFilePath = os.path.join(
                virtualServerDataDirectoryPath,
                self.applicationName + '.pickle')
            xmlFilePath = os.path.join(
                virtualServerDataDirectoryPath,
                self.applicationName + '.xml')
            rcFile = None
            if os.access(xmlFilePath, os.F_OK):
                rcFile = open(xmlFilePath, 'rb')
                import gnosis.xml.pickle as glasnostPickle
                glasnostPickle.setParanoia(0)
            elif os.access(pickleFilePath, os.F_OK):
                import cPickle as glasnostPickle
                rcFile = open(pickleFilePath, 'rb')
            if rcFile is None:
                virtualServer = self.VirtualServer()
                virtualServer.init(self, virtualServerId)
            else:
                fcntl.lockf(rcFile, fcntl.LOCK_SH)
                try:
                    versionLine = rcFile.readline(20).strip()
                    splittedVersion = versionLine.split('.')
                    version = 0.0
                    factor = 1000000.0
                    for number in splittedVersion:
                        version += int(number) * factor
                        factor = factor / 1000
                except ValueError:
                    version = 0
                    rcFile.seek(0)
                print 'file:',  rcFile
                data = glasnostPickle.load(rcFile)
                fcntl.lockf(rcFile, fcntl.LOCK_UN)
                rcFile.close()
                if isinstance(data, VirtualServer):
                    virtualServer = data
                else:
                    virtualServer = self.VirtualServer()
                    virtualServer.initFromOldData(self, virtualServerId, data)
                if version < 5004:
                    if not hasattr(virtualServer, 'virtualServerId') \
                       or not virtualServer.virtualServerId:
                        virtualServer.virtualServerId = \
                                                    virtualServer.applicationId
                        del virtualServer.applicationId
                virtualServer.check(self, virtualServerId)
                self.repairVirtualServer(virtualServer, version)
            virtualServer.dataDirectoryPath = virtualServerDataDirectoryPath
        else:
            virtualServer = self.VirtualServer()
            virtualServer.init(self, virtualServerId)
        return virtualServer

    def pullContext(self):
        context.pull()

    def pushContext(self, virtualServerId, clientToken, userToken):
        applicationToken = mainModule.applicationTokens[virtualServerId]
        dispatcherId = commonTools.extractDispatcherId(virtualServerId)
        serverRole = self.applicationRole
        applicationId = commonTools.makeApplicationId(dispatcherId, serverRole)
        if clientToken is None:
            clientToken = ''
        if userToken is None:
            userToken = ''
        context.push(
            applicationId = applicationId, # = virtualServerId
            applicationToken = applicationToken,
            clientToken = clientToken,
            dispatcherId = dispatcherId,
            server = self,
            serverRole = serverRole,
            userToken = userToken,
            )

    def registerToDispatcher(self):
        serverRole = self.applicationRole
        for dispatcherId in self.dispatcherIds:
            applicationId = commonTools.makeApplicationId(
                dispatcherId, serverRole)
            context.push(
                applicationId = applicationId, # = virtualServerId
                dispatcherId = dispatcherId,
                server = self,
                serverRole = serverRole,
                )
            mainModule.applicationTokens[applicationId] = getApplicationToken()
            registerServer(self.hostName, self.port)
            context.pull()

    def registerXmlRpcMethod(self, methodName, method = None):
        if method is None:
            method = getattr(self, methodName)
        self.xmlRpcServer.add_method(methodName, method)

    def registerXmlRpcMethods(self):
        self.registerXmlRpcMethod('addDispatcher', self.addDispatcherXmlRpc)
        self.registerXmlRpcMethod('canGetAdmin')

    def removeIdsFromVirtualServer(self, dispatcherId, rolesToKeep):
        virtualServerId = '%s/%s' % (dispatcherId, self.applicationRole)
        if not self.virtualServers.has_key(virtualServerId):
            print 'Missing virtual server for %s' % dispatcherId
            return 2
        self.virtualServers[virtualServerId].removeIds(rolesToKeep)
        self.saveVirtualServer(self.virtualServers[virtualServerId])
        return None

    def removeVirtualServerHistory(self, dispatcherId):
        return None

    def repairVirtualServer(self, virtualServer, version):
        pass

    def run(self):
        signal.signal(signal.SIGINT, self.handleSignal)
        signal.signal(signal.SIGHUP, self.handleSignal)
        signal.signal(signal.SIGTERM, self.handleSignal)
        if getConfig(self.applicationName, 'BypassXmlRpc') != 'true':
            while 1:
                try:
                    self.xmlRpcServer.serve_forever()
                except KeyboardInterrupt:
                    break
                except:
                    traceback.print_exc()
        else:
            while 1:
                try:
                    self.fastXmlRpcServer.serve_forever()
                except KeyboardInterrupt:
                    break
                except:
                    traceback.print_exc()

    def saveVirtualServer(self, virtualServer):
        if not self.useDataFile:
            return
        import cPickle as glasnostPickle
        virtualServerId = virtualServer.virtualServerId
        if virtualServerId is None:
            virtualServerDataDirectoryPath = self.dataDirectoryPath
        else:
            hostName = extractApplicationHostName(virtualServerId)
            virtualServerDataDirectoryPath = getConfig(
                hostName, 'DataDirectoryPath')
            if not virtualServerDataDirectoryPath:
                virtualServerDataDirectoryPath = getConfig(
                    'Misc', 'DispatcherDataDirectoryPath',
                    vars = {'hostname': hostName})
            if not os.access(virtualServerDataDirectoryPath, os.F_OK):
                os.mkdir(virtualServerDataDirectoryPath)
                os.chmod(virtualServerDataDirectoryPath, 0750)
        rcFilePath = os.path.join(
            virtualServerDataDirectoryPath,
            self.applicationName + '.pickle')
        rcFile = open(rcFilePath, 'wb')
        os.chmod(rcFilePath, 0640)
        fcntl.lockf(rcFile, fcntl.LOCK_EX)
        rcFile.write(self.version + '\n')
        virtualServer.lock.acquire()
        glasnostPickle.dump(virtualServer, rcFile, 1)
        virtualServer.lock.release()
        fcntl.lockf(rcFile, fcntl.LOCK_UN)
        rcFile.close()
        xmlFilePath = os.path.join(
            virtualServerDataDirectoryPath,
            self.applicationName + '.xml')
        if os.access(xmlFilePath, os.F_OK):
            os.remove(xmlFilePath)

    def setDispatcherIds(self):
        self.dispatcherIds = getConfig(
           'Misc', 'DispatcherURLs', default = 'glasnost://localhost')
        self.dispatcherIds = self.dispatcherIds.replace(',', ' ')
        self.dispatcherIds = self.dispatcherIds.strip().split()
        for i in range(len(self.dispatcherIds)):
            if self.dispatcherIds[i][-1] == '/':
                self.dispatcherIds[i] = self.dispatcherIds[i][:-1]
        # second dispatcher is system dispatcher
        self.dispatcherIds.insert(1, 'glasnost://system')

    def setLogFile(self):
        logFilePath = '%s/%s.log' % (
            getConfig('Misc', 'LogFilesDir', default='/var/log/glasnost'),
            self.applicationName)
        try:
            logFile = open(logFilePath, 'a', 1)
        except IOError:
            logFile = open('/dev/null', 'w')
        sys.stdout = sys.__stdout__ = logFile
        sys.stderr = sys.__stderr__ = logFile

    def startXmlRpcServer(self):
        self.hostName = getConfig(self.applicationName, 'ServerHostName')
        self.port = eval(getConfig(self.applicationName, 'ServerPort'), {})
        if self.isThreaded:
            self.xmlRpcServer = xmlrpcThreadingServer(('', self.port))
        else:
            self.xmlRpcServer = xmlrpcServer(('', self.port))
        self.registerXmlRpcMethods()

        if self.isThreaded:
            self.fastXmlRpcServer = xmlrpcFastThreadingServer(
                self.xmlRpcServer)
        else:
            self.fastXmlRpcServer = xmlrpcFastServer(self.xmlRpcServer)
        mainModule.applicationTokens = {}
        self.registerToDispatcher()

    def usage(self):
        print 'Usage: %s ### FIXME: To Do' % self.applicationName


class AdministrableServerMixin:
    adminClassName = None # To override.

    def canGetAdmin(self, virtualServerId, clientToken, userToken):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            dispatcherId = commonTools.extractDispatcherId(virtualServerId)
            result = dispatcherId != 'glasnost://system'
            return result
        finally:
            self.pullContext()

    def canModifyAdmin(self, virtualServerId, clientToken, userToken):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            result = self.isAdmin(virtualServerId, clientToken, userToken)
            return result
        finally:
            self.pullContext()

    def getAdminCore(self):
        virtualServerId = context.getVar('applicationId')
        virtualServer = self.getVirtualServer(virtualServerId)
        return virtualServer.admin

    def getAdminXmlRpc(self, virtualServerId, clientToken, userToken):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            admin = self.getAdminCore()
            admin.acquireNonCore()
            try:
                result = admin.exportToXmlRpc()
            finally:
                admin.releaseNonCore()
            return result
        finally:
            self.pullContext()

    def isAdmin(self, virtualServerId, clientToken, userToken):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            if not userToken:
                return 0
            virtualServer = self.getVirtualServer(virtualServerId)
            admin = virtualServer.admin
            if not admin.adminsSet:
                return 1
            return getProxyForServerRole('authentication').setContainsUser(
                admin.adminsSet)
        finally:
            self.pullContext()

    def modifyAdminXmlRpc(
        self, virtualServerId, clientToken, userToken, adminImport):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            if not self.isAdmin(virtualServerId, clientToken, userToken):
                raise faults.UserAccessDenied()
            adminChanges = self.newThing(
                'object', adminImport['__thingName__'])
            adminChanges.importFromXmlRpc(adminImport)
            virtualServer = self.getVirtualServer(virtualServerId)
            admin = virtualServer.admin
            admin.acquireNonCore()
            try:
                admin.checkModifyIsPossible(adminChanges)
                admin.modify(adminChanges)
                admin.saveNonCore()
            finally:
                admin.releaseNonCore()
            self.saveVirtualServer(virtualServer)
            #invalidateValue(Missing id in admin)
            return admin.version
        finally:
            self.pullContext()

    def newAdmin(self):
        return self.newThing('object', self.adminClassName)

    def registerXmlRpcMethods(self):
        self.registerXmlRpcMethod('canModifyAdmin')
        self.registerXmlRpcMethod('getAdmin', self.getAdminXmlRpc)
        self.registerXmlRpcMethod('isAdmin')
        self.registerXmlRpcMethod('modifyAdmin', self.modifyAdminXmlRpc)


class ObjectsServer(AdministrableServerMixin, Server):
    VirtualServer = ObjectsVirtualServer
    dataDirectoryPath = None
    objectClassName = None # To override.
    useAdminReadersSet = 1
    useAdminWritersSet = 1

    def addObjectXmlRpc(self, virtualServerId, clientToken, userToken,
                        objectImport):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            virtualServer = self.getVirtualServer(virtualServerId)
            object = self.newThing('object', objectImport['__thingName__'])
            object.importFromXmlRpc(objectImport)

            # User must be an admin or (a writer and a writer of this object))
            if not self.canAddObject(virtualServerId, clientToken, userToken) \
                   and (
                not self.isAdmin(virtualServerId, clientToken, userToken)
                and not (
                self.useAdminWritersSet
                and getProxyForServerRole('authentication').setContainsUser(
                self.getAdminCore().writersSet)
                and (
                not object.isSlotName('writersSet')
                or getProxyForServerRole('authentication').setContainsUser(
                object.getSlot('writersSet').getValue())))):
                clientId = getApplicationId(clientToken)
                if not object.canBeCreatedByClient(clientId):
                    raise faults.UserAccessDenied()
            object.checkAddIsPossible()
            object.setAutomaticalSlots()
            object.id = self.getNextId()
            virtualServer.objects[object.id] = object
            object.saveNonCore()
            object.releaseNonCore()
            self.saveVirtualServer(virtualServer)
            return object.id
        finally:
            self.pullContext()
    addObjectXmlRpc.allowSubscriptionAfter = 1

    def canAddObject(self, virtualServerId, clientToken, userToken):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            result = self.isAdmin(virtualServerId, clientToken, userToken) or (
                self.useAdminWritersSet
                and getProxyForServerRole('authentication').setContainsUser(
                self.getAdminCore().writersSet))
            return result
        finally:
            self.pullContext()

    def canDeleteObject(self, virtualServerId, clientToken, userToken,
                        objectId):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            virtualServer = self.getVirtualServer(virtualServerId)
            if not virtualServer.objects.has_key(objectId):
                return 0
            object = virtualServer.objects[objectId]
            return self.isAdmin(virtualServerId, clientToken, userToken) or (
                object.isSlotName('writersSet')
                and getProxyForServerRole('authentication').setContainsUser(
                object.getSlot('writersSet').getValue()))
        finally:
            self.pullContext()

    def canGetObject(self, virtualServerId, clientToken, userToken, objectId):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            virtualServer = self.getVirtualServer(virtualServerId)
            if not virtualServer.objects.has_key(objectId):
                return 0
            object = virtualServer.objects[objectId]
            if self.isAdmin(virtualServerId, clientToken, userToken):
                return 1
            clientId = getApplicationId(clientToken)
            if object.canBeGottenByClient(clientId):
                return 1
            admin = self.getAdminCore()
            if object.isSlotName('readersSet'):
                readersSet = object.getSlot('readersSet').getValue()
                return not readersSet or getProxyForServerRole(
                    'authentication').setContainsUser(readersSet)
            else:
                return not self.useAdminReadersSet or not admin.readersSet \
                       or getProxyForServerRole(
                    'authentication').setContainsUser(admin.readersSet)
        finally:
            self.pullContext()

    def canGetObjects(self, virtualServerId, clientToken, userToken):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            clientId = getApplicationId(clientToken)
            if self.isAdmin(virtualServerId, clientToken, userToken):
                return 1
            admin = self.getAdminCore()
            if (not self.useAdminReadersSet or not admin.readersSet or
                getProxyForServerRole('authentication').setContainsUser(
                admin.readersSet)):
                return 1
            return 0
        finally:
            self.pullContext()

    def canModifyObject(self, virtualServerId, clientToken, userToken,
                        objectId):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            virtualServer = self.getVirtualServer(virtualServerId)
            if not virtualServer.objects.has_key(objectId):
                return 0
            object = virtualServer.objects[objectId]
            if not object.canBeModified():
                return 0
            return self.isAdmin(virtualServerId, clientToken, userToken) or (
                object.isSlotName('writersSet')
                and getProxyForServerRole('authentication').setContainsUser(
                object.getSlot('writersSet').getValue()))
        finally:
            self.pullContext()

    def deleteObject(self, virtualServerId, clientToken, userToken, objectId):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            virtualServer = self.getVirtualServer(virtualServerId)
            clientId = getApplicationId(clientToken)
            if not virtualServer.objects.has_key(objectId):
                raise faults.MissingItem(objectId)
            object = virtualServer.objects[objectId]
            if not self.canDeleteObject(
                virtualServerId, clientToken, userToken, objectId):
                if not object.canBeDeletedByClient(clientId):
                    raise faults.UserAccessDenied()
            del virtualServer.objects[objectId]
            object.removeNonCore()
            self.saveVirtualServer(virtualServer)
            invalidateValue(objectId)
            return objectId # We have to return something.
        finally:
            self.pullContext()

    def getLastObjectIds(
        self, virtualServerId, clientToken, userToken, objectsCount,
        possibleReadersSet, possibleWritersSet):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            virtualServer = self.getVirtualServer(virtualServerId)
            isAdmin = self.isAdmin(virtualServerId, clientToken, userToken)
            possibleReaderIds = getSetContainedIds(
                possibleReadersSet, ['people'], serverId = virtualServerId)
            possibleWriterIds = getSetContainedIds(
                possibleWritersSet, ['people'], serverId = virtualServerId)
            objectIds = virtualServer.objects.keys()
            if objectIds and hasattr(
                virtualServer.objects[objectIds[0]], 'modificationTime'):
                def modificationTimeSorter(
                    xId, yId, virtualServer = virtualServer):
                    return cmp(virtualServer.objects[yId].modificationTime,
                               virtualServer.objects[xId].modificationTime)
                objectIds.sort(modificationTimeSorter)
            result = []
            for objectId in objectIds:
                object = virtualServer.objects[objectId]
                if object.isSlotName('readersSet'):
                    hasReadersSet = 1
                    readersSet = object.getSlot('readersSet').getValue()
                else:
                    hasReadersSet = 0
                if not isAdmin and hasReadersSet and readersSet \
                       and not getProxyForServerRole(
                    'authentication').setContainsUser(readersSet):
                    continue
                if not possibleReaderIds or not hasReadersSet \
                       or not readersSet:
                    readerFound = 1
                else:
                    readerFound = 0
                    objectReaderIds = getSetContainedIds(
                        readersSet, ['people'], serverId = virtualServerId)
                    if objectReaderIds:
                        for possibleReaderId in possibleReaderIds:
                            if possibleReaderId in objectReaderIds:
                                readerFound = 1
                                break
                if not possibleWriterIds \
                       or not object.isSlotName('writersSet'):
                    writerFound = 1
                else:
                    writerFound = 0
                    objectWriterIds = getSetContainedIds(
                        object.getSlot('writersSet').getValue(), ['people'],
                        serverId = virtualServerId)
                    if objectWriterIds:
                        for possibleWriterId in possibleWriterIds:
                            if possibleWriterId in objectWriterIds:
                                writerFound = 1
                                break
                if readerFound and writerFound:
                    result.append(objectId)
                    if not objectsCount == -1 and len(result) >= objectsCount:
                        break
            return result
        finally:
            self.pullContext()

    def getNextId(self):
        virtualServerId = context.getVar('applicationId')
        virtualServer = self.getVirtualServer(virtualServerId)
        virtualServer.lock.acquire()
        objectId = virtualServer.getNextLocalId()
        virtualServer.lock.release()
        return '%s/%s' % (virtualServerId, objectId)

    # The real XML-RPC function is getObjectXmlRpc.
    def getObjectCore(self, objectId):
        virtualServerId = self.computeVirtualServerId(objectId)
        virtualServer = self.getVirtualServer(virtualServerId)
        if not virtualServer.objects.has_key(objectId):
            raise faults.MissingItem(objectId)
        object = virtualServer.objects[objectId]
        clientToken = context.getVar('clientToken')
        userToken = context.getVar('userToken')
        if not self.canGetObject(
            virtualServerId, clientToken, userToken, objectId):
            raise faults.UserAccessDenied()
        return object

    def getObjectIds(self, virtualServerId, clientToken, userToken):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            if not self.canGetObjects(virtualServerId, clientToken, userToken):
                raise faults.UserAccessDenied()
            virtualServer = self.getVirtualServer(virtualServerId)
            admin = self.getAdminCore()
            isAdmin = self.isAdmin(virtualServerId, clientToken, userToken)
            clientId = getApplicationId(clientToken)
            result = []
            for objectId, objectCore in virtualServer.objects.items():
                if self.canGetObject(
                    virtualServerId, clientToken, userToken, objectId) or \
                    objectCore.canBeGottenByClient(clientId):
                    result.append(objectId)
            return result
        finally:
            self.pullContext()

    def getObjectItemLabelsXmlRpc(self, virtualServerId, clientToken,
                                  userToken, objectIds):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            virtualServer = self.getVirtualServer(virtualServerId)
            labels = {}
            for objectId in objectIds:
                if not virtualServer.objects.has_key(objectId):
                    raise faults.MissingItem(objectId)
                object = virtualServer.objects[objectId]
                labels[objectId] =  utf8(object.getLabel())
            return labels
        finally:
            self.pullContext()

    def getObjectItemLabelXmlRpc(self, virtualServerId, clientToken, userToken,
                                 objectId):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            virtualServer = self.getVirtualServer(virtualServerId)
            if not virtualServer.objects.has_key(objectId):
                raise faults.MissingItem(objectId)
            object = virtualServer.objects[objectId]
            result = utf8(object.getLabel())
            return result
        finally:
            self.pullContext()

    def getObjectLabelAndLanguageXmlRpc(self, virtualServerId, clientToken,
                                        userToken, objectId):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            virtualServer = self.getVirtualServer(virtualServerId)
            if not virtualServer.objects.has_key(objectId):
                raise faults.MissingItem(objectId)
            object = virtualServer.objects[objectId]
            result = [utf8(object.getLabel()), object.getLabelLanguage()]
            return result
        finally:
            self.pullContext()

    def getObjectStringFromDigestXmlRpc(
        self, virtualServerId, clientToken, userToken, objectId, path, digest):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            path = iso8859_15(path)
            virtualServer = self.getVirtualServer(virtualServerId)
            if not virtualServer.objects.has_key(objectId):
                raise faults.MissingItem(objectId)
            object = virtualServer.objects[objectId]
            object.acquireNonCore()
            try:
                # The objectLabel is accessible from everybody.
                if path == 'self.getLabel()':
                    result = getStringFromDigest(object.getLabel(), digest)
                else:
                    if not self.canGetObject(
                        virtualServerId, clientToken, userToken, objectId):
                        raise faults.UserAccessDenied()
                    try:
                        part = eval(path, {'self': object})
                    except:
                        raise faults.NonExistentSlotPath(path)
                    result = getStringFromDigest(part, digest)
            finally:
                object.releaseNonCore()
            if result is None:
                raise faults.UnknownStringDigest(digest)
            return utf8(result)
        finally:
            self.pullContext()

    def getObjectXmlRpc(self, virtualServerId, clientToken, userToken,
                        objectId):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            object = self.getObjectCore(objectId)
            object.acquireNonCore()
            try:
                result = object.exportToXmlRpc()
            finally:
                object.releaseNonCore()
            return result
        finally:
            self.pullContext()

    def getPartialObjectXmlRpc(self, virtualServerId, clientToken, userToken,
                               objectId, requiredSlotNames):
        if not requiredSlotNames:
            return self.getObjectXmlRpc(
                virtualServerId, clientToken, userToken, objectId)
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            object = self.getObjectCore(objectId)
            useNonCore = 0
            if not 'id' in requiredSlotNames \
               and object.isSlotName('id') \
               and object.getSlot('id').getKind().isPublicForXmlRpc:
                requiredSlotNames.append('id')
            if not 'language' in requiredSlotNames \
               and object.isSlotName('language') \
               and object.getSlot('language').getKind().isPublicForXmlRpc:
                requiredSlotNames.append('language')
            if not 'readersSet' in requiredSlotNames \
               and object.isSlotName('readersSet') \
               and object.getSlot('readersSet').getKind().isPublicForXmlRpc:
                requiredSlotNames.append('readersSet')
            if not 'version' in requiredSlotNames \
               and object.isSlotName('version') \
               and object.getSlot('version').getKind().isPublicForXmlRpc:
                requiredSlotNames.append('version')
            for slotName in object.getSlotNames():
                if not slotName in requiredSlotNames:
                    continue
                slot = object.getSlot(slotName)
                kind = slot.getKind()
                if not kind.isInCore(slot):
                    useNonCore = 1
                    break
            if useNonCore:
                object.acquireNonCore()
            try:
                result = object.exportToXmlRpc(
                    requiredSlotNames = requiredSlotNames)
            finally:
                if useNonCore:
                    object.releaseNonCore()
            return result
        finally:
            self.pullContext()

    def hasObject(self, virtualServerId, clientToken, userToken, objectId):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            virtualServer = self.getVirtualServer(virtualServerId)
            return virtualServer.objects.has_key(objectId)
        finally:
            self.pullContext()

    def modifyObjectXmlRpc(self, virtualServerId, clientToken, userToken,
                           objectImport):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            objectChanges = self.newThing(
                'object', objectImport['__thingName__'])
            objectChanges.importFromXmlRpc(objectImport)
            virtualServer = self.getVirtualServer(virtualServerId)
            if not virtualServer.objects.has_key(objectChanges.id):
                raise faults.MissingItem(objectChanges.id)
            object = virtualServer.objects[objectChanges.id]
            object.acquireNonCore()
            try:
                if not object.canBeModified():
                    raise faults.ReadOnlyObject()
                if not self.canModifyObject(
                    virtualServerId, clientToken, userToken, object.id) \
                    or not (
                    self.isAdmin(virtualServerId, clientToken, userToken)
                    or not objectChanges.isSlotName('writersSet')
                    or getProxyForServerRole('authentication').setContainsUser(
                    objectChanges.getSlot('writersSet').getValue())):
                    clientId = getApplicationId(clientToken)
                    if not object.canBeModifiedByClient(clientId):
                        raise faults.UserAccessDenied()
                object.checkModifyIsPossible(objectChanges)
                object.modify(objectChanges)
                object.saveNonCore()
            finally:
                object.releaseNonCore()
            self.saveVirtualServer(virtualServer)
            invalidateValue(object.id)
            return object.version
        finally:
            self.pullContext()

    def modifyPartialObjectXmlRpc(self, virtualServerId, clientToken,
                                  userToken, objectImport, givenSlotNames):
        if not givenSlotNames:
            return self.modifyObjectXmlRpc(
                virtualServerId, clientToken, userToken, objectImport)
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            if not objectImport.has_key('id'):
                raise faults.MissingId()
            id = objectImport['id']
            virtualServer = self.getVirtualServer(virtualServerId)
            if not virtualServer.objects.has_key(id):
                raise faults.MissingItem(id)
            object = virtualServer.objects[id]
            object.acquireNonCore()
            try:
                if not object.canBeModified():
                    raise faults.ReadOnlyObject()

                objectChanges = self.newThing(
                    'object', objectImport['__thingName__'])
                if not 'id' in givenSlotNames:
                    givenSlotNames.append('id')
                if not 'version' in givenSlotNames:
                    givenSlotNames.append('version')
                object.initPartial(
                    objectChanges, givenSlotNames = givenSlotNames)
                objectChanges.importFromXmlRpc(objectImport)

                if not self.canModifyObject(
                    virtualServerId, clientToken, userToken, object.id) \
                    or not (
                    self.isAdmin(virtualServerId, clientToken, userToken)
                    or not objectChanges.isSlotName('writersSet')
                    or getProxyForServerRole('authentication').setContainsUser(
                    objectChanges.getSlot('writersSet').getValue())):
                        clientId = getApplicationId(clientToken)
                        if not object.canBeModifiedByClient(clientId):
                            raise faults.UserAccessDenied()
                object.checkModifyIsPossible(
                    objectChanges, givenSlotNames = givenSlotNames)
                object.modify(
                    objectChanges, givenSlotNames = givenSlotNames)
                object.saveNonCore()
            finally:
                object.releaseNonCore()
            self.saveVirtualServer(virtualServer)
            invalidateValue(object.id)
            return object.version
        finally:
            self.pullContext()

    def newObject(self):
        return self.newThing('object', self.objectClassName)

    def registerXmlRpcMethods(self):
        Server.registerXmlRpcMethods(self)
        AdministrableServerMixin.registerXmlRpcMethods(self)
        self.registerXmlRpcMethod('addObject', self.addObjectXmlRpc)
        self.registerXmlRpcMethod('canAddObject')
        self.registerXmlRpcMethod('canDeleteObject')
        self.registerXmlRpcMethod('canGetObject')
        self.registerXmlRpcMethod('canGetObjects')
        self.registerXmlRpcMethod('canModifyObject')
        self.registerXmlRpcMethod('deleteObject')
        self.registerXmlRpcMethod('getLastObjectIds')
        self.registerXmlRpcMethod('getObject', self.getObjectXmlRpc)
        self.registerXmlRpcMethod('getObjectIds')
        self.registerXmlRpcMethod('getObjectItemLabel',
                                  self.getObjectItemLabelXmlRpc)
        self.registerXmlRpcMethod('getObjectItemLabels',
                                  self.getObjectItemLabelsXmlRpc)
        self.registerXmlRpcMethod('getObjectLabelAndLanguage',
                                  self.getObjectLabelAndLanguageXmlRpc)
        self.registerXmlRpcMethod('getObjectStringFromDigest',
                                  self.getObjectStringFromDigestXmlRpc)
        self.registerXmlRpcMethod('getPartialObject',
                                  self.getPartialObjectXmlRpc)
        self.registerXmlRpcMethod('hasObject')
        self.registerXmlRpcMethod('modifyObject', self.modifyObjectXmlRpc)
        self.registerXmlRpcMethod('modifyPartialObject',
                                  self.modifyPartialObjectXmlRpc)
        self.registerXmlRpcMethod('searchObjectIds')
        self.registerXmlRpcMethod('sortObjectIds')
        self.registerXmlRpcMethod('subscribe')
        self.registerXmlRpcMethod('unsubscribe')

    def searchObjectIds(
        self, virtualServerId, clientToken, userToken, searchDict):
        """ Search keywords syntax:
        'attribute_name_EQ = value' means attribute=value 
        'attribute_name_LT = value' means attribute < value 
        'attribute_name_GT = value' means attribute > value 
        'attribute_name_LE = value' means attribute <= value 
        'attribute_name_GE = value' means attribute >= value 
        'attribute_name_IN = value' means value contains attribute 
        'attribute_name_CT = value' means value in attribute
        'searchTerms = value' is for plain text blank-separated search terms
        
        """        
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            if not self.canGetObjects(virtualServerId, clientToken, userToken):
                raise faults.UserAccessDenied()
            
            virtualServer = self.getVirtualServer(virtualServerId)
            searchSlotNames = []
            searchOperators = []
            searchValues = []
            def testType(value, operator):
                return (operator in ['EQ', 'LT', 'GT', 'LE', 'GE'] and \
                       type(value) == type('') or \
                       operator == 'CT' or \
                       operator == 'IN' and type(value) == type([]) or \
                       operator == 'PT' and \
                        type(value) in [type(''), type([])])
            for sk in searchDict.keys():
                slotName = sk[:-3]
                searchOperator = sk[-2:]
                searchValue = searchDict[sk]
                if not testType(searchValue, searchOperator):
                    raise Exception(
                        'Invalid search value = %s' % str(searchValue))
                searchSlotNames.append(slotName)
                searchOperators.append(searchOperator)
                searchValues.append(searchValue) 
            clientId = getApplicationId(clientToken)
            object = self.newObject()
            allSlotNames = object.getSlotNames()
            useNonCore = 0
            for slotName in searchSlotNames:
                if not slotName in allSlotNames:
                    raise Exception('Search not valid')
                slot = object.getSlot(slotName)
                kind = slot.getKind()
                if not kind.isInCore(slot):
                    useNonCore = 1
                    break
            result = []
            def plainTextSearch(stringObject, searchValue):
                # FIXME this place is for a better plain text search
                for term in searchValue:
                    term = iso8859_15(term)
                    if stringObject.find(term) == -1:
                        return 0
                return 1
            def test(searchValue, testValue, operator):
                return (operator == 'EQ' and testValue == searchValue) or \
                       (operator == 'LT' and testValue < searchValue) or \
                       (operator == 'GT' and testValue > searchValue) or \
                       (operator == 'LE' and testValue <= searchValue) or \
                       (operator == 'GE' and testValue >= searchValue) or \
                       (operator == 'CT' and searchValue in testValue) or \
                       (operator == 'IN' and testValue in searchValue)
            for objectId, object in virtualServer.objects.items():
                if self.canGetObject(
                    virtualServerId, clientToken, userToken, objectId) or \
                    object.canBeGottenByClient(clientId):
                    if useNonCore:
                        object.acquireNonCore()
                    kept = 1
                    for i in range(len(searchSlotNames)):
                        slotName = searchSlotNames[i]
                        searchOperator = searchOperators[i]
                        searchValue = searchValues[i]
                        slot = object.getSlot(slotName)
                        testValue = slot.getValue()
                        if searchOperator == 'PT' and \
                               not plainTextSearch(testValue, searchValue):
                            kept = 0
                            break
                        if searchOperator in [
                            'EQ', 'LT', 'LE', 'GT', 'GE', 'CT', 'IN'] and \
                               not test(
                            searchValue, testValue, searchOperator):
                            kept = 0
                            break
                    if kept:
                        result.append(objectId)
            return result
        finally:
            self.pullContext()

    def sortObjectIds(self, virtualServerId, clientToken, userToken,
                      objectIds, sortRule): 
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            virtualServer = self.getVirtualServer(virtualServerId)
            if not objectIds:
                objectIds = virtualServer.objects.keys()
            else:
                objectIds = [ objectId
                              for objectId in objectIds
                              if virtualServer.objects.has_key(objectId) ]
            def straightIdSorter(xId, yId):
                try:
                    xParts = '/'.split(string.atoi(xId))
                    yParts = '/'.split(string.atoi(yId))
                    return cmp(xParts[-2], yParts[-2]) or \
                           cmp(xParts[-1], yParts[-1])
                except:
                    raise Exception(
                        'Unable to compare what looks like 2 invalid Glasnost'
                        ' identifiers')
            def straightLabelSorter(xId, yId, virtualServer = virtualServer):
                return cmp(virtualServer.objects[xId].getLabel(),
                           virtualServer.objects[yId].getLabel())
            if sortRule == 'id':
                try:
                    objectIds.sort(straightIdSorter)
                except:
                    raise Exception('Unable to compare objects ids')
            elif sortRule == 'modificationTime':
                try:
                    objects = {}
                    for objectId in objectIds:
                        objects[objectId] = virtualServer.objects[objectId]
                    slot = objects.values()[0].getSlot('modificationTime')
                    kind = slot.getKind()
                    if not kind.isInCore(slot):
                        map (lambda x:x.acquireNonCore(), objects.values())
                    def reverseModificationTimeSorter(
                        xId, yId, objects = objects):
                        return cmp(
                            objects[yId].modificationTime,
                            objects[xId].modificationTime)
                    objectIds.sort(reverseModificationTimeSorter)
                except:
                    raise Exception(
                        'Unable to compare objects modification time')
            elif sortRule == 'date':
                try:
                    objects = {}
                    for objectId in objectIds:
                        objects[objectId] = virtualServer.objects[objectId]
                    slot = objects.values()[0].getSlot('date')
                    kind = slot.getKind()
                    if not kind.isInCore(slot):
                        map (lambda x:x.acquireNonCore(), objects.values())
                    def reverseDateSorter(
                        xId, yId, objects = objects):
                        return cmp(
                            objects[yId].date, objects[xId].date)
                    objectIds.sort(reverseDateSorter)
                except:
                    raise Exception('Unable to compare objects date')
            else:
                objectIds.sort(straightLabelSorter)
            return objectIds
        finally:
            self.pullContext()

    def subscribe(self, virtualServerId, clientToken, userToken, method, when,
                  callBackName):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            if not self.xmlRpcServer._methods.has_key(method) \
               and not method == '_ALL_':
                raise Exception('No such method')
            if not when in ('before', 'after'):
                raise Exception('Wrong value for "when" (= %s)' % when)

            isAdmin = self.isAdmin(virtualServerId, clientToken, userToken)
            if not isAdmin:
                raise faults.UserAccessDenied()

            clientId = getApplicationId(clientToken)

            self.xmlRpcServer.add_subscription(
                when, virtualServerId, method, clientId, callBackName)
            return 0 # we have to return something
        finally:
            self.pullContext()

    def unsubscribe(self, virtualServerId, clientToken, userToken, method,
                    when, callBackName):
        self.pushContext(virtualServerId, clientToken, userToken)
        try:
            isAdmin = self.isAdmin(virtualServerId, clientToken, userToken)
            if not isAdmin:
                raise faults.UserAccessDenied()

            clientId = getApplicationId(clientToken)

            self.xmlRpcServer.remove_subscription(
                when, virtualServerId, method, clientId, callBackName)
            return 0 # we have to return something
        finally:
            self.pullContext()

