# xmlrpcServer.py - An xmlrpc server framework for python
# Copyright (C) 2001 by Ken McIvor.  All rights reserved.
#
# Contains works derived from xmlrpcserver.py (a simple XML-RPC server for
# Python), Copyright (C) 1999 by Fredrik Lundh, Copyright (C) 1999 by Secret 
# Labs AB.  xmlrpcserver.py is available as part of version 0.9.8 of 
# PythonWare's XML-RPC library, found at
# http://www.pythonware.com/downloads/index.htm
#
# Contains works derived from xmlrpc_registry.py (A method registry for use with
# xmlrpclib), Copyright (C) 2001 by Eric Kidd (all rights reserved).
# xmlrpc_registry.py is available at
# http://xmlrpc-c.sourceforge.net/hacks/xmlrpc_registry.py
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission. 
#  
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

# Changelog
# 
# 9/5/2001 Initial (1.0) release
# 6/15/2002 Adapted to Glasnost (Faults are encapsulated) (manou)

# TODO
# 
# - Comment, comment, comment.  Docstrings are my friend...
# - Add signal handling framework for INT, TERM, HUP
# - Move exception handling from do_POST to dispatch_call
# - Add __init__ keyword arg to disable logging
# - Add support for servers using unix domain sockets for better performance

import sys
import os
import time
import string
import socket
import threading
import traceback
import SocketServer
import BaseHTTPServer
import xmlrpclib

import glasnost.common.context as context
import glasnost.common.faults as faults
from glasnost.common import XmlRpcTools

from glasnost.proxy.DispatcherProxy import callServer

try:
    from M2Crypto import SSL
except ImportError:
    SSL = None

__version__ = '1.0'


# Some type names for use in method signatures.
INT="int"
BOOLEAN="boolean"
DOUBLE="double"
STRING="string"
DATETIME="dateTime.iso8601"
BASE64="base64"
ARRAY="array"
STRUCT="struct"


# Some error codes, borrowed from xmlrpc-c.
INTERNAL_ERROR = -500
TYPE_ERROR = -501
INDEX_ERROR = -502
PARSE_ERROR = -503
NETWORK_ERROR = -504
TIMEOUT_ERROR = -505
NO_SUCH_METHOD_ERROR = -506
REQUEST_REFUSED_ERROR = -507
INTROSPECTION_DISABLED_ERROR = -508
LIMIT_EXCEEDED_ERROR = -509
INVALID_UTF8_ERROR = -510


class notificationThread(threading.Thread):
    def __init__(self, subscriberId, callBack, params):
        threading.Thread.__init__(self)
        self.subscriberId = subscriberId
        self.callBack = callBack
        self.params = params

    def run(self):
        try:
            callServer(self.subscriberId, self.callBack, self.params)
        except:
            pass # notification failed. Who should we warn?


class xmlrpcServer(SocketServer.TCPServer):
    __server_version = 'xmlrpcServer/' + __version__
    __python_version = 'Python ' + string.split(sys.version)[0]
    monthname = [None,
        'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
        'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
    allow_reuse_address = 1

    def __init__ (self, server_address, allow_introspection=1, sslContext=None,
                  handler=None):
        if sslContext and SSL:
            self.__class__.__bases__ = (SSL.SSLServer,)
            self.finish = self.sslFinish

        if handler == None:
               handler = xmlrpcRequestHandler
    
        i = 0
        while 1:
            try:
                if sslContext and SSL:
                    SSL.SSLServer.__init__(
                        self, server_address, handler, sslContext)
                else:
                    SocketServer.TCPServer.__init__(
                        self, server_address, handler)
                break
            except socket.error, e:
                if e.args[0] == 98:
                    i += 1
                    if i == 5:
                        raise
                    time.sleep(2)
                    continue
                raise

        if server_address[0] == '':
            host = socket.gethostbyaddr(socket.gethostname())[0]
        else:
            host = server_address[0]
        self._address = '%s:%s' % (host, server_address[1])
        self._allow_introspection = allow_introspection
        self._methods = {}
        self._signatures = {}
        self._help = {}
        self._capabilities = {}
        self._default_method = None
        self._subscriptions_before = {}
        self._subscriptions_after = {}
        self.log_message('%s starting up', self.version_string())
        if allow_introspection:
            state = 'enabled'
        else:
            state = 'disabled'
        self.log_message('introspection is %s', state)
        self._install_system_methods()

    def sslFinish(self):
        print 'finishing request'
        self.request.set_shutdown(
            SSL.SSL_RECEIVED_SHUTDOWN | SSL.SSL_SENT_SHUTDOWN)
        print 'almost done'
        self.request.close()
        print 'done'

    def serve_forever(self):
        self.log_message('awaiting connections')
        SocketServer.TCPServer.serve_forever(self)

    def _install_system_methods(self):
        self.add_method('system.listMethods',
                        self.system_listMethods,
                        [[ARRAY]])
        self.add_method('system.methodSignature',
                        self.system_methodSignature,
                        [[ARRAY, STRING]])
        self.add_method('system.methodHelp',
                        self.system_methodHelp,
                        [[STRING, STRING]])
        self.add_method('system.multicall',
                        self.system_multicall,
                        [[ARRAY, ARRAY]])
        self.add_method('system.getCapabilities',
                        self.system_getCapabilities,
                        [[STRUCT]])

    def log_date_time_string(self):
        now = time.time()
        year, month, day, hh, mm, ss, x, y, z = time.localtime(now)
        s = "%02d/%3s/%04d %02d:%02d:%02d" % ( day, self.monthname[month], 
            year, hh, mm, ss)
        return s

    def log_error(self, format, *args):
        self.write_error("%s [%s] %s" % (self._address,
            self.log_date_time_string(), format % args))

    def log_message(self, format, *args):
        self.write_message("%s [%s] %s" % (self._address, \
            self.log_date_time_string(), format % args))

    def write_error(self, message):
        sys.stderr.write("ERROR: %s\n" % message)

    def write_message(self, message):
        sys.stderr.write("LOG: %s\n" % message)

    def add_method(self, name, method, signature=None, help=None):
        if help == None:
            help = ''
        if signature == None:
            signature = 'undef'

        self._methods[name] = method
        self._signatures[name] = signature
        self._help[name] = help
        self.log_message('add_method %s', name)

    def add_capability(self, name, specUrl, specVersion):
        self._capabilities[name] = {
            'specUrl': str(specUrl),
            'specVersion': int(specVersion)
        }
        self.log_message('add_capability %s %s %s', name, specUrl, 
            specVersion)

    def add_subscription(self, when, virtualServerId, functionName,
                         subscriberId, callBackName):
        if when == 'after':
            subscriptions = self._subscriptions_after
        else:
            subscriptions = self._subscriptions_before
            
        listSubscriptions = subscriptions.setdefault(virtualServerId, [])
        listSubscriptions.append( (functionName, subscriberId, callBackName) )

    def remove_subscription(self, when, virtualServerId, functionName,
                            subscriberId, callBackName):
        if when == 'any':
            self.remove_subscription(
                'before', virtualServerId, functionName, subscriberId,
                callBackName)
            self.remove_subscription(
                'after', virtualServerId, functionName, subscriberId,
                callBackName)
            return

        if when == 'after':
            subscriptions = self._subscriptions_after
        else:
            subscriptions = self._subscriptions_before
        listSubscriptions = subscriptions.setdefault(virtualServerId, [])
        subscriptions[virtualServerId] = [
            x
            for x in listSubscriptions
            if x[1] != subscriberId or (
                (x[0] != functionName and functionName != '_ALL_') \
                and (x[2] != callBackName and callBackName != 'any'))]
        
    def set_default_method(self, method):
        self._default_method = method

    def version_string(self):
        return self.__server_version + ' (' + self.__python_version + ')'

    def dispatch_call(self, name, params, isDirectCall = 0):
        if self._methods.has_key(name):
            method = self._methods[name]
        else:
            method = self._default_method
        if method == None:
            self._no_such_method(name)
        try:
            if params:
                self.send_before_notify(params[0], name, params)
            if not isDirectCall:
                context.clear()
            ## FIXME: when all server the use auto deserialization suppress
            ## the condition
            autoConvert = (hasattr(method.im_class, 'useAutoDeserialize')
                           and method.im_class.useAutoDeserialize) \
                           or (hasattr(method, 'useAutoDeserialize')
                               and method.useAutoDeserialize)
            if autoConvert:
                # FIXME: A quick hack. It assumes that each method declared
                # with autoConvert set, always begins with the three parameters
                # virtualServerId, clientToken & userToken.
                # This is mandatory, because sometimes, the method
                # convertFromXmlRpc requires this context variables.
                # If this hack works well, the pushContext and pullContext
                # inside each "autoConverted" method should be removed.
                server = method.im_self
                assert len(params) >= 3
                virtualServerId, clientToken, userToken = params[:3]
                server.pushContext(virtualServerId, clientToken, userToken)
                try:
                    params = XmlRpcTools.convertFromXmlRpc(params)
                    result = apply(method, params)
                    result = XmlRpcTools.convertToXmlRpc(result)
                finally:
                    server.pullContext()
            else:
                result = apply(method, params)
            if not isDirectCall:
                assert context.isClear()
            if params:
                self.send_after_notify(params[0], name, result)
            return result
        except faults.Fault:
            self.log_error(
                'error for request %s(%s) on %s with autoConvert %s',
                name, params, method, autoConvert)
            traceback.print_exc()
            raise
        except:
            self.log_error(
                'error for request %s(%s) on %s with autoConvert %s',
                name, params, method, autoConvert)
            traceback.print_exc()
            raise faults.UnknownServerException()
        

    def _no_such_method(self, name):
        raise xmlrpclib.Fault(NO_SUCH_METHOD_ERROR, str(name))

    def _introspection_check(self):
        if not self._allow_introspection:
            raise xmlrpclib.Fault(INTROSPECTION_DISABLED_ERROR,
                ("Introspection has been disabled on this server."))
    
    def send_before_notify(self, virtualServerId, functionName, params):
        if functionName.startswith('system.'):
            return
        subscriptions = self._subscriptions_before
        
        for name, subscriber, callBack in subscriptions.setdefault(
            virtualServerId, []):
            if name not in (functionName, '_ALL_'):
                continue
            self.send_notify(
                subscriber, callBack, [virtualServerId, functionName, params])

    def send_after_notify(self, virtualServerId, functionName, result):
        if functionName.startswith('system.'):
            return
        subscriptions = self._subscriptions_after
        for name, subscriber, callBack in subscriptions.setdefault(
            virtualServerId, []):
            if name not in (functionName, '_ALL_'):
                continue
            self.send_notify(
                subscriber, callBack, [virtualServerId, functionName, result])

    def send_notify(self, subscriberId, callBack, params):
        notifThread = notificationThread(subscriberId, callBack, params)
        notifThread.start()
        
    def system_listMethods(self):
        self._introspection_check()
        return self._methods.keys()

    def system_methodSignature(self, name):
        self._introspection_check()
        if self._signatures.has_key(name):
            return self._signatures[name]
        else:
            self._no_such_method(name)

    def system_methodHelp(self, name):
        self._introspection_check()
        if self._help.has_key(name):
            return self._help[name]
        else:
            self._no_such_method(name)

    def system_getCapabilities(self):
        self._introspection_check()
        return self._capabilities

    def system_multicall(self, calls):
        results = []
        for call in calls:
            # XXX: individual faults and exceptions are not logged
            # probable solution: move exception handling from do_POST to
            # dispatch_call
            try:
                name = call['methodName']
                params = call['params']
                if name == 'system.multicall':
                    errmsg = "Recursive system.multicall forbidden"
                    raise xmlrpclib.Fault(REQUEST_REFUSED_ERROR, errmsg)
                print '  multicall, calling', name
                result = [self.dispatch_call(name, params)]
            except xmlrpclib.Fault, fault:
                result = {'faultCode': fault.faultCode,
                          'faultString': fault.faultString}
            except:
                info = sys.exc_info()
                errmsg = "%s:%s" % (info[0], info[1])
                result = {'faultCode': 1, 'faultString': errmsg}
            results.append(result)
        return results


class xmlrpcThreadingServer(SocketServer.ThreadingMixIn, xmlrpcServer):
    pass


class xmlrpcForkingServer(SocketServer.ForkingMixIn, xmlrpcServer):
    pass


class xmlrpcRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    def __init__(self, request, client_address, server):
        BaseHTTPServer.BaseHTTPRequestHandler.__init__(self, request, \
            client_address, server)
        
        self.server_version = 'xmlrpcServer/' + __version__

    def log_rpc_request(self, method):
        self.log_message(
            'end of request %s %s', self.client_address[0], method)

    def log_rpc_fault(self, method, fault):
        self.log_error('fault %s %s %s "%s"', self.client_address[0],
            method, fault.faultCode, fault.faultString)

    def log_exception(self, method, info):
        tb = traceback.extract_tb(info[2])
        self.log_error('exception %s %s "%s" %s %s %s',
            self.client_address[0],
            method,
            string.strip(traceback.format_exception_only(info[0], info[1])[0]),
            tb[-1][0], tb[-1][1], tb[-1][2])

    def log_error(self, format, *args):
        self.server.write_error("%s [%s] %s" % (self.address_string(),
            self.log_date_time_string(), format % args))

    def log_message(self, format, *args):
        self.server.write_message("%s [%s] %s" % (self.address_string(), \
            self.log_date_time_string(), format % args))

    def log_request(self, code='-', size='-'):
        pass

    def do_POST(self):
        data = self.rfile.read(int(self.headers["content-length"]))
        params, method = xmlrpclib.loads(data)
        self.log_message(
            'begin of request %s %s', self.client_address[0], method)
        try:
            response = self.server.dispatch_call(method, params)
            if type(response) != type(()):
                response = (response,)
        except xmlrpclib.Fault, e:
            self.log_rpc_fault(method, e)
            response = xmlrpclib.dumps(e)
        except:
            info = sys.exc_info()
            self.log_exception(method, info)
            response = xmlrpclib.dumps(xmlrpclib.Fault(1, "%s" %
                string.strip(traceback.format_exception_only(info[0],
                info[1])[0])))
        else:
            try:
                response = xmlrpclib.dumps( response, methodresponse=1)
            except:
                info = sys.exc_info()
                self.log_exception(method, info)
                response = xmlrpclib.dumps(xmlrpclib.Fault(1, "%s" %
                    string.strip(traceback.format_exception_only(info[0],
                    info[1])[0])))
            self.log_rpc_request(method)

        self.send_response(200)
        self.send_header("Content-type", "text/xml")
        self.send_header("Content-length", str(len(response)))
        self.end_headers()
        self.wfile.write(response)
        self.wfile.flush()
        self.connection.shutdown(1)

    def version_string(self):
        return self.server.version_string()


import marshal


class xmlRpcFastRequestHandler(SocketServer.StreamRequestHandler,
                               xmlrpcRequestHandler):
    def __init__(self, request, client_address, server):
        self.server = server
        self.client_address = client_address
        SocketServer.BaseRequestHandler.__init__(
            self, request, client_address, server)

    def address_string(self):
        return self.client_address
        
    def handle(self):
        bytes = self.rfile.readline()
        data = marshal.loads(self.rfile.read(int(bytes)))
        params, method = data['params'], data['methodName']
        self.log_message('begin of fast request: %s', method)
        try:
            response = self.server.dispatch_call(method, params)
            if type(response) != type(()):
                response = (response,)
        except xmlrpclib.Fault, e:
            response = [e.faultCode, e.faultString]
        except:
            info = sys.exc_info()
            e = xmlrpclib.Fault(1, "%s" %
                    string.strip(traceback.format_exception_only(info[0],
                    info[1])[0]))
            response = [e.faultCode, e.faultString]
        response = marshal.dumps(response)
        self.wfile.write('%11d\n' % len(response))
        self.wfile.write(response)
        self.wfile.flush()
        #self.wfile.close()


class xmlrpcFastServer(SocketServer.UnixStreamServer, xmlrpcServer):
    def __init__(self, server):
        for k,v in server.__dict__.items():
            setattr(self, k, v)
        handler = xmlRpcFastRequestHandler
        server_address = '/tmp/.glasnost:%s' \
                         % self._address[self._address.rfind(':')+1:]
        self._address = server_address
        try:
            os.unlink(server_address)
        except OSError:
            pass
        SocketServer.UnixStreamServer.__init__(self, server_address, handler)
        os.chmod(server_address, 0777)

    def serve_forever(self):
        self.log_message('awaiting connections (fast)')
        SocketServer.UnixStreamServer.serve_forever(self)
        self.log_message('finished serving (fast)')


class xmlrpcFastForkingServer(SocketServer.ForkingMixIn, xmlrpcFastServer):
    pass


class xmlrpcFastThreadingServer(SocketServer.ThreadingMixIn, xmlrpcFastServer):
    pass


def _main():
    if len(sys.argv) > 1:
        addr = ('', int(sys.argv[1]))
    else:    
        addr = ('', 8000)
    serv = xmlrpcServer(addr)
    serv.add_capability('xmlrpc', 'http://www.xmlrpc.com/spec', 1)
    serv.serve_forever()


if __name__ == '__main__':
    _main()
